"""

Large parts of this are just python translations of lib2geom.

"""


import collections, math



class XYZ(collections.namedtuple('XYZ','x y z')):
    def __neg__(self):
        return XYZ(-self.x,-self.y,-self.z)
    
    def __add__(self, other):
        return XYZ(self.x+other.x,self.y+other.y,self.z+other.z)

    def __sub__(self, other):
        return XYZ(self.x-other.x,self.y-other.y,self.z-other.z)

    def __mul__(self, other):
        return XYZ(self.x * other, self.y * other, self.z * other)
        
    def __rmul__(self, other):
        return self*other

    def dot(self, other):
        return self.x*other.x+self.y*other.y+self.z*other.z
    
    def cross(self, other):
        return XYZ(
            self.y*other.z-self.z*other.y,
            self.z*other.x-self.x*other.z,
            self.x*other.y-self.y*other.x
            )
    
    def mag2(self):
        return self.dot(self)
    
    def mag(self):
        return math.sqrt(self.mag2())

    def unit(self):
        return (1.0/self.mag()) * self



class Linear(collections.namedtuple('Linear','a0 a1')):
    """ f(t) = a0*(1-t)+a1*t
    """

    def __call__(self, t):
        return self.a0*(1-t)+self.a1*t

    def __add__(self, other): 
        return type(self)(self.a0+other.a0,self.a1+other.a1)

    def __sub__(self, other): 
        return type(self)(self.a0-other.a0,self.a1-other.a1)

    def __mul__(self, other): 
        return type(self)(self.a0*other,self.a1*other)
        
    def __rmul__(self, other):
        return self*other

    def tri(self):
        return self.a1-self.a0

    def hat(self):
        return (self.a1+self.a0)*0.5


class S_basis(tuple):
    """ length>=1 immutable list of Linear
    
        f(t) = sum self[i](t) * s**t where s = t*(1-t)
    """

    def __call__(self, t):
        s = t*(1-t)
        result = self[0](t)
        p = s
        for i in xrange(1,len(self)):
            result += self[i](t) * p
            p *= s
        return result
    
    def __repr__(self):
        return 'S_basis(' + repr(list(self)) + ')'
        
    def scaled(self,other): 
        return type(self)( item*other for item in self )
    
    def shifted(self, i):
        return type(self)((self[0]*0,)*i+tuple(self))
    
    def truncated(self, n):
        return type(self)(self[:n])
    
    def _compat(self,other):
        size = max(len(self),len(other))
        if len(self) < size:
            self = type(other)(tuple(self) + (self[0]*0,)*(size-len(self)))
        if len(other) < size:
            other = type(other)(tuple(other) + (other[0]*0,)*(size-len(other)))
        return size, self, other
    
    def __add__(self, other):
        size, self, other = self._compat(other)
        return type(self)( self[i]+other[i] for i in xrange(size) )

    def __sub__(self, other):
        size, self, other = self._compat(other)
        return type(self)( self[i]-other[i] for i in xrange(size) )

    def multiplied(self, other, operator):
        zero = operator(self[0].a0*0,other[0].a0*0)
        
        c = [Linear(zero,zero)]*(len(self)+len(other))
        for j in xrange(len(other)):
            for i in xrange(j,j+len(self)):
                tri = operator(other[j].tri(),self[i-j].tri())
                c[i+1] = c[i+1] + Linear(-tri,-tri)
                c[i] = c[i] + Linear(operator(other[j].a0,self[i-j].a0), operator(other[j].a1,self[i-j].a1))
        #while len(c) > 1 and c[-1] == zero:
        #    del c[-1]
        return S_basis(c)
    
    def __mul__(self, other):
        return self.multiplied(other, lambda a,b: a*b)
        
    def __rmul__(self, other):
        return self*other
    
    def dot(self, other):
        return self.multiplied(other, lambda a,b: a.dot(b))
    
    def divided(self, other, k):
        return least_squares(
            S_basis([Linear(self[0].a0/other[0].a0,self[0].a1/other[0].a1)]),
            lambda x: other*x-self,
            lambda x: other,
            lambda x: ZERO,
            k)
    
        #remainder = self
        #result = [ ]
        #for i in xrange(k):
        #    if len(remainder) <= i:
        #        break
        #    ci = Linear(remainder[i].a0/other[0].a0, remainder[i].a1/other[0].a1)
        #    result.append(ci)
        #    remainder = remainder - (S_basis([ci])*other).shifted(i)
        #return S_basis(result)

    def reciprocal(self, k):
        return ONE.divided(self, k)
    
    def derivative(self):
        c = [ ]
        for k in xrange(len(self)-1):
             d = (2*k+1)*(self[k].a1 - self[k].a0)
             c.append(Linear(
                 d + (k+1)*self[k+1].a0,
                 d - (k+1)*self[k+1].a1
                 ))
        k = len(self)-1
        d = (2*k+1)*(self[k].a1 - self[k].a0)
        c.append(Linear(d,d))
        return S_basis(c)

    def integral(self):
        a = [ self[0]*0 ]
        for k in xrange(1,len(self)+1):
            ahat = self[k-1].tri()*(-1.0/(2*k))
            a.append(Linear(ahat,ahat))
        aTri = self[0].a0*0
        for k in xrange(len(self)-1,-1,-1):
            aTri = (self[k].hat() + (k+1)*0.5*aTri)*(1.0/(2*k+1))
            a[k] = a[k] + Linear(-0.5*aTri,0.5*aTri)
        return S_basis(a)
    
    def sqrt(self, k):
        return least_squares(self, lambda x: x*x-self, lambda x: x.scaled(2), lambda x: ONE.scaled(2), k)

        #""" Calculate square root by newton's method """
        #result = self
        #for i in xrange(iters):
        #    result = (result+self.divided(result, k)).scaled(0.5)
        #return result

    def compose(self, other):
        """ Calculate f(t)=self(other(t)) """
        s = (ONE-other)*other
        result = S_basis([self[0]*0])
        for i in xrange(len(self)-1,-1,-1):
            result = result*s + (ONE-other).scaled(self[i].a0) + other.scaled(self[i].a1)
            #S_basis([Linear(self[i].a0,self[i].a0)]) + other.scaled(self[i].a1-self[i].a0)
        return result

    def solve(self, target, k, iters=20):
        """ Solve self.compose(x) = target for x using Newton's method            
        """
        result = target
        deriv = self.derivative()
        for i in xrange(iters):
            result = result - (self.compose(result) - target).divided(deriv.compose(result), k)
            result = result.truncated(k)
        return result.truncated(k)

    def inverse(self, k, iters=5):
        return self.solve(IDENTITY, k, iters)

        
ONE = S_basis([Linear(1.0,1.0)])
ZERO = S_basis([Linear(0.0,0.0)])
IDENTITY = S_basis([Linear(0.0,1.0)])


#def fit(initial, transform, k, iters):
#    def score(guess):
#        tx = transform(guess)
#        return (tx*tx).integral()[0].tri()
#    
#    result = initial
#    current = score(result)
#    
#    #basis = [ ]
#    #slant = IDENTITY-ONE.scaled(0.5)
#    #for i in xrange(k):
#    #    basis.append(ONE.shifted(i))
#    #    basis.append(slant.shifted(i))
#        #basis.append(IDENTITY.shifted(i))
#        #basis.append((ONE-IDENTITY).shifted(i))
#        
#    #needs to be an orthogonal basis
#    
#    # Legendre polynomials
#    X = IDENTITY.scaled(2.0)-ONE   # Compress from -1,1 to 0,1
#    basis = [ ONE, X ]
#    while len(basis) < k:
#        n = len(basis)
#        basis.append( (basis[-1]*X).scaled((2*n+1)/(n+1.0)) - basis[-2].scaled(n/(n+1.0)) )
#    
#    #plot(*basis)
#    #foo
#    
#    step = 1.0
#    for i in xrange(iters):
#        for item in basis:
#            low = result + item.scaled(-step)
#            low_score = score(low)
#            high = result + item.scaled(step)
#            high_score = score(high)
#            
#            c = current
#            a = (high_score+low_score)*0.5-current
#            b = high_score-a-c
#
#            if high_score < current:
#                result = high
#                current = high_score
#                #step *= 2.0
#            if low_score < current:
#                result = low
#                current = low_score
#                #step *= 2.0
#
#            if a:
#                min_point = -0.5/step*b/a
#                new = result + item.scaled( min_point )
#                new_score = score(new)                
#                if new_score < current:
#                    result = new
#                    current = new_score
#                    step = max(step, abs(min_point)*2.0)
#        step *= 0.85
#        print current, step
#        #if not step: break
#
#    return result


def newtonoid(initial, fp,fpp, k):
    """ Choose x to minimize the integral of f(x) over [0,1] """
    def scorep(a,b):
        return (b*fp(a)).integral()[0].tri()    
    def scorepp(a,b):
        return (b*b*fpp(a)).integral()[0].tri()

    result = initial
    #current = score(result)
    
    # Legendre polynomials
    X = IDENTITY.scaled(2.0)-ONE   # Compress from -1,1 to 0,1
    basis = [ ONE, X ]
    while len(basis) < k*2:
        n = len(basis)
        basis.append( (basis[-1]*X).scaled((2*n+1)/(n+1.0)) - basis[-2].scaled(n/(n+1.0)) )
    
    #plot(*basis)
    #foo
    
    step = 1.0
    for i in xrange(k*8):
        for item in basis:
            #c = score(result)
            #b = scorep(result)
            #a = 0.5*scorepp(result)            
            #step = -b/2a
            
            step = -scorep(result,item)/scorepp(result,item)
            #print step
            new = result + item.scaled(step)
        
                #new = result + item.scaled( min_point )
            #new_score = score(new)                
            #if new_score < current:
            result = new
            #current = new_score
        #step *= 0.85
        #print current, step
        #if not step: break

    return result

def least_squares(guess, f,fp,fpp, k):
    """ Choose x to minimize the integral of f(x)^2 over [0,1] """    
    #def f2(x): 
    #    y = f(x)
    #    return y*y
    def f2p(x):
        return (fp(x)*f(x)).scaled(2.0)
    def f2pp(x):
        yp = fp(x)
        return (fpp(x)*f(x)+yp*yp).scaled(2.0)
    return newtonoid(
        guess,
        f2p,
        f2pp,
        k)

"""TODO: optimize path for aesthetics
"""

class Frame(collections.namedtuple('Frame', 'origin x y z')):
    """ Origin and orthogonal basis. """

    def apply(self, point):
        return self.origin + self.x*point.x + self.y*point.y+ self.z*point.z
    
    def unapply(self, point):
        point = point - self.origin
        return XYZ(self.x.dot(point),self.y.dot(point),self.z.dot(point))


class Path(collections.namedtuple('Path','path velocity normal position')):
    def find(self, position):
        low = 0.0
        high = 1.0
        for i in xrange(32):
            mid = (low+high)*0.5
            value = self.position(mid)
            if position < value:
                high = mid
            elif position > value:
                low = mid
            else:
                return mid #Unlikely.
        return (low+high)*0.5

    def get_length(self):
        return self.position[0].a1

    def get_frame(self, position):
        t = self.find(position)
        point = self.path(t)
        
        z = self.velocity(t).unit()
        x = self.normal(t)
        x = (x - z*x.dot(z)).unit()
        y = z.cross(x)
        return Frame(point, x,y,z)

    def get_point(self, position):
        return self.path(self.find(position))
    
    def get_bentness(self, a,b):
        aa = self.find(a)
        bb = self.find(b)
        seg = self.path.compose(S_basis([Linear(aa,bb)]))
        straight = seg[:1]
        diff = seg-straight
        return math.sqrt( diff.dot(diff).integral()[0].tri() / seg[0].tri().mag2() )
        

def path(point0,vec0,norm0,point1,vec1,norm1):
    #a = S_basis([Linear(XYZ(0.0,0.0,0.0), XYZ(1.0,1.0,0.0))])
    #b = S_basis([Linear(XYZ(3.0,0.0,0.0), XYZ(0.0,-1.0,0.0))])
    #arc = a + (S_basis([Linear(-a[0].tri(),a[0].tri())])+b).shifted(1)
    tri = point1-point0
    
    length = tri.mag()
    print '.', length
    vec0 = vec0.unit()
    vec1 = vec1.unit()
    for i in xrange(4):
        s = length
        path = S_basis([Linear(point0,point1),Linear(vec0*s-tri,vec1*-s+tri)])    
        velocity = path.derivative()
        speed = velocity.dot(velocity).sqrt( 6 )
        position = speed.integral()
        position = position - ONE.scaled(position[0].a0)
        length = position[0].a1
        print '-', length
    
    normal = S_basis([Linear(norm0,norm1)])
    
    return Path(path, velocity, normal, position)


def plot(*items):
    import pylab

    ts = [ i/100.0 for i in xrange(101) ]
    for item in items:
        pylab.plot(ts, [ item(t) for t in ts ])
    pylab.show()
         
if __name__ == '__main__':
    import pylab
    
    x = IDENTITY+ONE.scaled(0.0)
    #dent = x.sqrt(2) #( (x).compose(IDENTITY) - IDENTITY ).divided(x.derivative(),5) #inverse(10,1)
    
    #dent = fit(x, lambda y: y*y-x, 10, 100)
    dent = least_squares(x, lambda y: y*y-x, lambda y: y.scaled(2), lambda y: ONE.scaled(2), 2)
    
    #dent = (dent*dent)#.truncated(20)
    #dent = b
    #r = 0.25
    #a = S_basis([Linear(0.5-r, 0.5+r)])
    #b = a.inverse(1)
    
    
    
    #a = ONE + IDENTITY*(ONE-IDENTITY)
    #a = a.compose(a)
    #b = a.reciprocal(5)
    #a = b.inverse(5)
    #b = a.inverse(5)
    
    #dent = x.compose(a).reciprocal(3).compose(b)
    #dent = b
    #dent = dent + (x-dent.reciprocal(5)).reciprocal(5)
    
    #a = S_basis([Linear(XYZ(0.0,0.0,0.0), XYZ(1.0,1.0,0.0))])
    #b = S_basis([Linear(XYZ(3.0,0.0,0.0), XYZ(0.0,-1.0,0.0))])
    #arc = a + (S_basis([Linear(-a[0].tri(),a[0].tri())])+b).shifted(1)
    #
    #velocity = arc.derivative()
    #speed = velocity.dot(velocity).sqrt(10)
    #position = speed.integral()
    #position = position - ONE.scaled(position(0.0))
    #scaler = position.scaled(1.0/position(1.0))
    #iscaler = scaler.inverse(10, 100)
    #dent = scaler.compose(iscaler)
    #
    #arc = arc.compose(scaler)
    
    #print position(0.0), position(1.0)
    
    #print arc.derivative()(0), arc.derivative()(1)
    
    ts = [ i/100.0 for i in xrange(101) ]
    #pylab.plot([ arc(t).x for t in ts ],[ arc(t).y for t in ts ],'.')
    pylab.plot([ dent(t) for t in ts ])
    pylab.plot([ x(t)**0.5 for t in ts ], 'o')
    pylab.show()
    #print arc(0.5)
    
    #print ONE
    #print ONE.compose(IDENTITY)
    
    #a = IDENTITY*IDENTITY + IDENTITY
    #b = a.inverse(10)
    #print a
    #print b
    #print a.compose(b).truncated(5)

    #x = S_basis([Linear(1.0,2.0)])
    #print (x*x)(3)
    #y = (x*x).sqrt(4)
    #for i in xrange(11):
    #    t = i/10.0
    #    print x(t), y(t)
    #
    #print S_basis([Linear(0.01,4.0)]).sqrt(4)







# Utilities for piecewise cornu representation of curves

from math import *

import clothoid
import cornu

class Segment:
    def __init__(self, z0, z1, th0, th1):
        self.z0 = z0
        self.z1 = z1
        self.th0 = th0
        self.th1 = th1
        self.compute()
    def __repr__(self):
        return '[' + `self.z0` + `self.z1` + ' ' + `self.th0` + ' ' + `self.th1` + ']'
    def compute(self):
        dx = self.z1[0] - self.z0[0]
        dy = self.z1[1] - self.z0[1]
        chord = hypot(dy, dx)
        chth = atan2(dy, dx)
        k0, k1 = clothoid.solve_clothoid(self.th0, self.th1)
        charc = clothoid.compute_chord(k0, k1)

        self.chord = chord
        self.chth = chth
        self.k0, self.k1 = k0, k1
        self.charc = charc
        self.arclen = chord / charc
        self.thmid = self.chth - self.th0 + 0.5 * self.k0 - 0.125 * self.k1

        self.setup_xy_fresnel()

    def setup_xy_fresnel(self):
        k0, k1 = self.k0, self.k1
        if k1 == 0: k1 = 1e-6 # hack
        if k1 != 0:
            sqrk1 = sqrt(2 * abs(k1))
            t0 = (k0 - .5 * k1) / sqrk1
            t1 = (k0 + .5 * k1) / sqrk1
            (y0, x0) = cornu.eval_cornu(t0)
            (y1, x1) = cornu.eval_cornu(t1)
            chord_th = atan2(y1 - y0, x1 - x0)
            chord = hypot(y1 - y0, x1 - x0)
            scale = self.chord / chord
            if k1 >= 0:
                th = self.chth - chord_th
                self.mxx = scale * cos(th)
                self.myx = scale * sin(th)
                self.mxy = -self.myx
                self.myy = self.mxx
            else:
                th = self.chth + chord_th
                self.mxx = scale * cos(th)
                self.myx = scale * sin(th)
                self.mxy = self.myx
                self.myy = -self.mxx
                # rotate -chord_th, flip top/bottom, rotate self.chth
            self.x0 = self.z0[0] - (self.mxx * x0 + self.mxy * y0)
            self.y0 = self.z0[1] - (self.myx * x0 + self.myy * y0)

    def th(self, s):
        u = s / self.arclen - 0.5
        return self.thmid + (0.5 * self.k1 * u + self.k0) * u

    def xy(self, s):
        # using fresnel integrals; polynomial approx might be better
        u = s / self.arclen - 0.5
        k0, k1 = self.k0, self.k1
        if k1 == 0: k1 = 1e-6 # hack
        if k1 != 0:
            sqrk1 = sqrt(2 * abs(k1))
            t = (k0 + u * k1) / sqrk1
            (y, x) = cornu.eval_cornu(t)
            return [self.x0 + self.mxx * x + self.mxy * y,
                    self.y0 + self.myx * x + self.myy * y]

    def find_extrema(self):
        # find solutions of th(s) = 0 mod pi/2
        # todo: find extra solutions when there's an inflection
        th0 = self.thmid + 0.125 * self.k1 - 0.5 * self.k0
        th1 = self.thmid + 0.125 * self.k1 + 0.5 * self.k0
        twooverpi = 2 / pi
        n0 = int(floor(th0 * twooverpi))
        n1 = int(floor(th1 * twooverpi))
        if th1 > th0: signum = 1
        else: signum = -1
        result = []
        for i in range(n0, n1, signum):
            th = pi/2 * (i + 0.5 * (signum + 1))
            a = .5 * self.k1
            b = self.k0
            c = self.thmid - th
            if a == 0:
                u1 = -c/b
                u2 = 1000
            else:
                sqrtdiscrim = sqrt(b * b - 4 * a * c)
                u1 = (-b - sqrtdiscrim) / (2 * a)
                u2 = (-b + sqrtdiscrim) / (2 * a)
            if u1 >= -0.5 and u1 < 0.5:
                result.append(self.arclen * (u1 + 0.5))
            if u2 >= -0.5 and u2 < 0.5:
                result.append(self.arclen * (u2 + 0.5))
        return result

class Curve:
    def __init__(self, segs):
        self.segs = segs
        self.compute()
    def compute(self):
        arclen = 0
        sstarts = []
        for seg in self.segs:
            sstarts.append(arclen)
            arclen += seg.arclen

        self.arclen = arclen
        self.sstarts = sstarts
    def th(self, s, deltas = False):
        u = s / self.arclen
        s = self.arclen * (u - floor(u))
        if s == 0 and not deltas: s = self.arclen
        i = 0
        while i < len(self.segs) - 1:
            # binary search would make a lot of sense here
            snext = self.sstarts[i + 1] 
            if s < snext or (not deltas and s == snext):
                break
            i += 1
        return self.segs[i].th(s - self.sstarts[i])
    def xy(self, s):
        u = s / self.arclen
        s = self.arclen * (u - floor(u))
        i = 0
        while i < len(self.segs) - 1:
            # binary search would make a lot of sense here
            if s <= self.sstarts[i + 1]:
                break
            i += 1
        return self.segs[i].xy(s - self.sstarts[i])
    def find_extrema(self):
        result = []
        for i in range(len(self.segs)):
            seg = self.segs[i]
            for s in seg.find_extrema():
                result.append(s + self.sstarts[i])
        return result
    def find_breaks(self):
        result = []
        for i in range(len(self.segs)):
            pseg = self.segs[(i + len(self.segs) - 1) % len(self.segs)]
            seg = self.segs[i]
            th = clothoid.mod_2pi(pseg.chth + pseg.th1 - (seg.chth - seg.th0))
            print '% pseg', pseg.chth + pseg.th1, 'seg', seg.chth - seg.th0
            pisline = pseg.k0 == 0 and pseg.k1 == 0
            sisline = seg.k0 == 0 and seg.k1 == 0
            if fabs(th) > 1e-3 or (pisline and not sisline) or (sisline and not pisline):
                result.append(self.sstarts[i])
        return result




import sys, os, math
sys.path.insert(0, os.path.split(__file__)[0])

import design, make, design_flute, profile, shape, pack

from nesoni import config

@config.help("""\
Produce 3D models using the output of "demakein design-*-flute:"
""")
@config.Bool_flag('open', 'Open both ends, thus requiring a cork.')
@config.Bool_flag('mill', 'Create shapes for milling (rather than 3D printing).')
@config.Float_flag('mill_diameter', 'Milling: Bit diameter for milling (affects gap size when packing pieces).')
@config.Float_flag('mill_length', 'Milling: Wood length for milling.')
@config.Float_flag('mill_width', 'Milling: Wood width for milling.')
@config.Float_flag('mill_thickness', 'Milling: Wood thickness for milling.')
@config.Int_flag('mill_scheme', 'Milling: Division scheme for milling.\nValid values: 2 3 4')
@config.Bool_flag('decorate', 'Add some decorations')
class Make_flute(make.Make_instrument):
    gap = 0.35

    mill = False
    open = False

    mill_diameter = 3.0    
    mill_length = 180.0
    mill_width = 130.0
    mill_thickness = 19.0
    mill_scheme = 4
    
    decorate = False
    
    SCHEMES = {
        2 : [
            [ 0.0, 0.6, 1.0 ],
            [ 0.0, 0.4, 1.0 ],
        ],
        
        3 : [
            [ 0.0, 0.3, 0.6, 1.0 ],
            [ 0.0, 0.4, 0.7, 1.0 ],
        ],
        
        4 : [
            [ 0.0, 0.2, 0.45, 0.7, 1.0 ],
            [ 0.0, 0.3, 0.55, 0.8, 1.0 ],
        ],
    }
    
    def run(self):
        working = self.working
        designer = working.designer
        spec = working.spec
        workspace = self.get_workspace()
    
        length = spec.length * 1.05   # Extend a bit to allow cork.
        
        if self.open:
            inner_profile = spec.inner.clipped(-50,length+50)
            
            cork_length = length - spec.length
            cork_diameter = spec.inner(spec.length)
            with open(workspace/(self.prefix+'cork.txt'),'wt') as f:
                f.write('Cork length %.1fmm\n' % cork_length)
                f.write('Cork diameter %.1fmm\n' % cork_diameter)
        else:
            inner_profile = spec.inner.clipped(-50,spec.length)
        
        width = max(spec.outer.low)
        
        outer_profile = spec.outer.clipped(0,length)
        if self.decorate:
           emfrac = 1.0-spec.hole_positions[-1]/length
           for frac, align in [ (0.0,1.0), (1.0-emfrac*2,1.0), (1.0,-1.0) ]:
               dpos = length * frac
               damount = outer_profile(dpos)*0.2
               dpos += damount * align
               deco_profile = profile.Profile(
                   [ dpos+damount*i for i in [-1,0,1]],
                   [ damount*i      for i in [0,1,0] ],
               )
               outer_profile = outer_profile + deco_profile


        self.make_instrument(
             inner_profile=inner_profile, outer_profile=outer_profile, 
             hole_positions=spec.hole_positions, hole_diameters=spec.hole_diameters, hole_vert_angles=spec.hole_angles, 
             hole_horiz_angles=[0.0]*7,
             xpad=[0.0]*7, 
             ypad=[0.0]*6+[0.75], 
             with_fingerpad=[True]*6+[False])
        
        #instrument = shape.extrude_profile(outer_profile)
        #outside = shape.extrude_profile(outer_profile)
        #
        #if self.open:        
        #    bore = shape.extrude_profile(spec.inner.clipped(-50,length+50))
        #else:
        #    bore = shape.extrude_profile(spec.inner.clipped(-50,spec.length))
        #
        ##xpad = [ .25,.25,0,0,0,0 ] + [ 0.0 ]
        ##ypad = [ .25,.25,0,0,0,0 ] + [ 0.75 ]
        #xpad = [ 0,0,0,0,0,0 ] + [ 0.0 ]
        #ypad = [ 0,0,0,0,0,0 ] + [ 0.75 ]
        #
        #
        #
        #for i, pos in enumerate(spec.hole_positions):
        #    angle = spec.hole_angles[i]
        #    radians = angle*math.pi/180.0
        #
        #    height = spec.outer(pos)*0.5 
        #    shift = math.sin(radians) * height
        #    
        #    hole_length = (
        #        math.sqrt(height*height+shift*shift) + 
        #        spec.hole_diameters[i]*0.5*abs(math.sin(radians)) + 
        #        4.0
        #    )
        #    hole = shape.prism(
        #        hole_length, spec.hole_diameters[i],
        #        shape.squared_circle(xpad[i], ypad[i]).with_effective_diameter
        #    )
        #    hole.rotate(1,0,0, -90-angle)
        #    hole.move(0,0,pos + shift)
        #    bore.add(hole)
        #    if angle:
        #        outside.remove(hole)
        #
        #instrument.remove(bore)    
        #instrument.rotate(0,0,1, 180)
        #self.save(instrument,'instrument')
        
        if self.mill:        
            shapes = pack.cut_and_pack(
                working.outside, working.bore,
                self.SCHEMES[self.mill_scheme][0], self.SCHEMES[self.mill_scheme][1],
                xsize=self.mill_length, 
                ysize=self.mill_width, 
                zsize=self.mill_thickness,
                bit_diameter=self.mill_diameter,
                save=self.save,
            )
        
        else:
            cut1 = min(spec.hole_positions[3],spec.inner_hole_positions[3])-spec.hole_diameters[3] * 0.75
            cut1 -= spec.inner(cut1)*0.5
            #cut1 = spec.inner_hole_positions[2]*0.5+spec.inner_hole_positions[3] * 0.5
            #cut2 = cut1 + (length-cut1)*0.3 + spec.outer.maximum()*0.7
            cut2 = length * 0.62

            self.segment([ cut1, cut2 ], length, up=False)
            
            #shapes = [ ]
            #
            #cuts = [ 
            #    cut1,
            #    cut2,
            #]
            #
            #remainder = working.instrument.copy()
            #for cut in cuts:
            #    d1 = spec.inner(cut)
            #    d3 = spec.outer(cut)
            #    d2 = (d1+d3) / 2.0
            #    
            #    
            #    d4 = spec.outer.maximum() * 2.0
            #    p1 = cut-d2*0.4
            #    p3 = cut+d2*0.4
            #    p2 = p1 + (d3-d1)*0.25
            #
            #    #0.4mm clearance all around
            #    # radius -> diameter    * 2
            #    # adjust both diamerers / 2
            #    # net effect            * 1
            #    # 0.4mm was loose
            #    
            #    d1a = d1 - self.gap
            #    p1b = p1 - self.gap
            #    
            #    d2a = d2 - self.gap
            #    d2b = d2 + self.gap
            #    
            #    prof_inside = profile.Profile(
            #        [ p1,  p2,  p3,  length+50 ],
            #        [ d1a, d2a, d2a, d4 ],
            #        [ d1a, d2a, d4,  d4 ],
            #    )
            #    prof_outside = profile.Profile(
            #        [ p1b, p2,  p3,  length+50 ],
            #        [ d1,  d2b, d2b, d4 ],
            #        [ d1,  d2b, d4,  d4 ],
            #    )
            #    mask_inside = shape.extrude_profile(prof_inside)
            #    mask_outside = shape.extrude_profile(prof_outside)
            #    
            #    item = remainder.copy()
            #    item.remove(mask_outside)                
            #    remainder.clip(mask_inside)
            #    shapes.append(item)
            #shapes.append(remainder)
            #shapes = shapes[::-1]
            #
            #for i, item in enumerate(shapes):
            #    item.rotate(0,1,0, 180)
            #
            #    self.save(item, self.prefix + 'segment-%d-of-%d' % (i+1,len(shapes)))
        
        #if self.forms == 1:
        #    lower, upper = shape.make_formwork(
        #        outside, bore, length,
        #        [ 0.0, 0.6, 1.0 ],
        #        [ 0.0, 0.4, 1.0 ],
        #        3.0, 6.0, self.thickness, 200.0
        #    )
        #    
        #    self.save(lower,'lower.stl')
        #    self.save(upper,'upper.stl')
        #    
        #    shape.show_only(instrument, lower, upper)
        #
        #elif self.forms == 2:
        #    lower1, upper1 = shape.make_formwork(
        #        outside, bore, length,
        #        [ 0.0, 0.36, 0.66, 1.0 ],
        #        [ ],
        #        3.0, 6.0, self.thickness, 200.0
        #    )
        #    lower2, upper2 = shape.make_formwork(
        #        outside, bore, length,
        #        [ ],
        #        [ 0.0, 0.25, 0.5, 0.75, 1.0 ],
        #        3.0, 6.0, self.thickness, 200.0
        #    )
        #    
        #    self.save(lower1,'lower1.stl')
        #    self.save(upper1,'upper1.stl')
        #    self.save(lower2,'lower2.stl')
        #    self.save(upper2,'upper2.stl')
        #    
        #    shape.show_only(instrument, lower1,upper1,lower2,upper2)
        #
        #else:
        #     assert False, 'Unsupported number of forms'
    

if __name__ == '__main__': 
    shape.main_action(Make_flute())




from math import *
import array
import sys
import random

import numarray as N
import numarray.linear_algebra as la

def eps_prologue(x0, y0, x1, y1, draw_box = False):
    print '%!PS-Adobe-3.0 EPSF'
    print '%%BoundingBox:', x0, y0, x1, y1
    print '%%EndComments'
    print '%%EndProlog'
    print '%%Page: 1 1'
    if draw_box:
        print x0, y0, 'moveto', x0, y1, 'lineto', x1, y1, 'lineto', x1, y0, 'lineto closepath stroke'

def eps_trailer():
    print '%%EOF'

# One step of 4th-order Runge-Kutta numerical integration - update y in place
def rk4(y, dydx, x, h, derivs):
    hh = h * .5
    h6 = h * (1./6)
    xh = x + hh
    yt = []
    for i in range(len(y)):
	yt.append(y[i] + hh * dydx[i])
    dyt = derivs(xh, yt)
    for i in range(len(y)):
	yt[i] = y[i] + hh * dyt[i]
    dym = derivs(xh, yt)
    for i in range(len(y)):
	yt[i] = y[i] + h * dym[i]
	dym[i] += dyt[i]
    dyt = derivs(x + h, yt)
    for i in range(len(y)):
	y[i] += h6 * (dydx[i] + dyt[i] + 2 * dym[i])

def run_elastica_half(sp, k0, lam1, lam2, n):
    def mec_derivs(x, ys):
	th, k = ys[2], ys[3]
	dx, dy = cos(th), sin(th)
	return [dx, dy, k, lam1 * dx + lam2 * dy, k * k]
    ys = [0, 0, 0, k0, 0]
    xyk = [(ys[0], ys[1], ys[3])]
    n = max(1, int(sp * n))
    h = float(sp) / n
    s = 0
    for i in range(n):
	dydx = mec_derivs(s, ys)
	rk4(ys, dydx, s, h, mec_derivs)
	xyk.append((ys[0], ys[1], ys[3]))
    return xyk, ys[2], ys[4]

def run_elastica(sm, sp, k0, lam1, lam2, n = 100):
    xykm, thm, costm = run_elastica_half(-sm, k0, -lam1, lam2, n)
    xykp, thp, costp = run_elastica_half(sp, k0, lam1, lam2, n)
    xyk = []
    for i in range(1, len(xykm)):
	x, y, k = xykm[i]
	xyk.append((-x, y, k))
    xyk.reverse()
    xyk.extend(xykp)
    x = xyk[-1][0] - xyk[0][0]
    y = xyk[-1][1] - xyk[0][1]
    th = thm + thp
    sth, cth = sin(thm), cos(thm)
    x, y = x * cth - y * sth, x * sth + y * cth
    result = []
    maxk = 0
    for xt, yt, kt in xyk:
        maxk = max(maxk, kt)
        result.append((xt, yt, kt))
    cost = costm + costp
    return result, cost, x, y, th

def run_mec_cos_pos(k, lam1, lam2, n = 1000):
    cost = 0
    th = 0
    x = 0
    y = 0
    dt = 1.0 / n
    result = [(0, 0)]
    for i in range(n):
        k1 = lam1 * cos(th) + lam2 * sin(th)

        cost += dt * k * k
        x += dt * cos(th)
        y += dt * sin(th)
        th += dt * k

        k += dt * k1
        result.append((x, y))
    return result, cost, x, y, th

def run_mec_cos(k, lam1, lam2, n = 1000):
    resp, costp, xp, yp, thp = run_mec_cos_pos(k*.5, lam1*.25, lam2*.25, n)
    resm, costm, xm, ym, thm = run_mec_cos_pos(k*.5, lam1*-.25, lam2*.25, n)
    cost = (costp + costm) * .5
    x, y = xp + xm, yp - ym
    th = thp + thm
    sth, cth = .5 * sin(thm), .5 * cos(thm)
    x, y = x * cth - y * sth, x * sth + y * cth
    result = []
    for i in range(1, n):
        xt, yt = resm[n - i - 1]
        result.append((-xt * cth - yt * sth, -xt * sth + yt * cth))
    for i in range(n):
        xt, yt = resp[i]
        result.append((xt * cth - yt * sth, xt * sth + yt * cth))
    return result, cost, x, y, th

def cross_prod(a, b):
    return [a[1] * b[2] - a[2] * b[1],
            a[2] * b[0] - a[0] * b[2],
            a[0] * b[1] - a[1] * b[0]]

def solve_mec(constraint_fnl):
    delta = 1e-3
    params = [pi, 0, 0]
    for i in range(20):
        k, lam1, lam2 = params
        xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
        #print i * .05, 'setgray'
        #plot(xys)
        c1c, c2c, costc = constraint_fnl(cost, x, y, th)
        print '% constraint_fnl =', c1c, c2c, 'cost =', costc

        dc1s = []
        dc2s = []
        for j in range(len(params)):
            params1 = N.array(params)
            params1[j] += delta
            k, lam1, lam2 = params1
            xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
            c1p, c2p, costp = constraint_fnl(cost, x, y, th)
            params1 = N.array(params)
            params1[j] -= delta
            k, lam1, lam2 = params1
            xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
            c1m, c2m, costm = constraint_fnl(cost, x, y, th)
            dc1s.append((c1p - c1m) / (2 * delta))
            dc2s.append((c2p - c2m) / (2 * delta))
        xp = cross_prod(dc1s, dc2s)
        xp = N.divide(xp, sqrt(N.dot(xp, xp))) # Normalize to unit length

        print '% dc1s =', dc1s
        print '% dc2s =', dc2s
        print '% xp =', xp
        
        # Compute second derivative wrt orthogonal vec
        params1 = N.array(params)
        for j in range(len(params)):
            params1[j] += delta * xp[j]
        k, lam1, lam2 = params1
        xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
        c1p, c2p, costp = constraint_fnl(cost, x, y, th)
        print '% constraint_fnl+ =', c1p, c2p, 'cost =', costp
        params1 = N.array(params)
        for j in range(len(params)):
            params1[j] -= delta * xp[j]
        k, lam1, lam2 = params1
        xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
        c1m, c2m, costm = constraint_fnl(cost, x, y, th)
        print '% constraint_fnl- =', c1m, c2m, 'cost =', costm
        d2cost = (costp + costm - 2 * costc) / (delta * delta)
        dcost = (costp - costm) / (2 * delta)

        print '% dcost =', dcost, 'd2cost =', d2cost
        if d2cost < 0: d2cost = .1
        # Make Jacobian matrix to invert
        jm = N.array([dc1s, dc2s, [x * d2cost for x in xp]])
        #print jm
        ji = la.inverse(jm)
        #print ji

        dp = N.dot(ji, [c1c, c2c, dcost])
        print '% dp =', dp
        print '% [right]=', [c1c, c2c, dcost]
        params -= dp * .1
        print '%', params
        sys.stdout.flush()
    return params

def solve_mec_3constr(constraint_fnl, n = 30, initparams = None):
    delta = 1e-3
    if initparams:
        params = N.array(initparams)
    else:
        params = [3.14, 0, 0]
    for i in range(n):
        k, lam1, lam2 = params
        xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
        c1c, c2c, c3c = constraint_fnl(cost, x, y, th)
        print '% constraint_fnl =', c1c, c2c, c3c

        dc1s = []
        dc2s = []
        dc3s = []
        for j in range(len(params)):
            params1 = N.array(params)
            params1[j] += delta
            k, lam1, lam2 = params1
            xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
            c1p, c2p, c3p = constraint_fnl(cost, x, y, th)
            params1 = N.array(params)
            params1[j] -= delta
            k, lam1, lam2 = params1
            xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
            c1m, c2m, c3m = constraint_fnl(cost, x, y, th)
            dc1s.append((c1p - c1m) / (2 * delta))
            dc2s.append((c2p - c2m) / (2 * delta))
            dc3s.append((c3p - c3m) / (2 * delta))

        # Make Jacobian matrix to invert
        jm = N.array([dc1s, dc2s, dc3s])
        #print jm
        ji = la.inverse(jm)

        dp = N.dot(ji, [c1c, c2c, c3c])
        if i < n/2: scale = .25
        else: scale = 1
        params -= scale * dp
        print '%', params
    return params

def mk_ths_fnl(th0, th1):
    def fnl(cost, x, y, th):
        actual_th0 = atan2(y, x)
        actual_th1 = th - actual_th0
        print '% x =', x, 'y =', y, 'hypot =', hypot(x, y)
        return th0 - actual_th0, th1 - actual_th1, cost
    return fnl

def mk_y_fnl(th0, th1, ytarg):
    def fnl(cost, x, y, th):
        actual_th0 = atan2(y, x)
        actual_th1 = th - actual_th0
        return th0 - actual_th0, th1 - actual_th1, ytarg - hypot(x, y)
    return fnl

def solve_mec_nested_inner(th0, th1, y, fnl):
    innerfnl = mk_y_fnl(th0, th1, y)
    params = [th0 + th1, 1e-6, 1e-6]
    params = solve_mec_3constr(innerfnl, 10, params)
    k, lam1, lam2 = params
    xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2, 100)
    c1, c2, c3 = fnl(cost, x, y, th)
    return c3, params

# Solve (th0, th1) mec to minimize fnl; use solve_mec_3constr as inner loop
# Use golden section search
def solve_mec_nested(th0, th1, fnl):
    R = .61803399
    C = 1 - R
    ax, cx = .6, .85
    bx = .5 * (ax + cx)

    x0, x3 = ax, cx
    if abs(cx - bx) > abs(bx - ax):
        x1, x2 = bx, bx + C * (cx - bx)
    else:
        x1, x2 = bx - C * (bx - ax), bx
    f1, p = solve_mec_nested_inner(th0, th1, x1, fnl)
    f2, p = solve_mec_nested_inner(th0, th1, x2, fnl)
    for i in range(10):
        print '%', x0, x1, x2, x3, ':', f1, f2
        if f2 < f1:
            x0, x1, x2 = x1, x2, R * x2 + C * x3
            f1 = f2
            f2, p = solve_mec_nested_inner(th0, th1, x2, fnl)
        else:
            x1, x2, x3 = R * x1 + C * x0, x1, x2
            f2 = f1
            f1, p = solve_mec_nested_inner(th0, th1, x1, fnl)
    if f1 < f2:
        x = x1
    else:
        x = x2
    cc, p = solve_mec_nested_inner(th0, th1, x, fnl)
    return p

def plot(xys):
    cmd = 'moveto'
    for xy in xys:
        print 306 + 200 * xy[0], 396 - 200 * xy[1], cmd
        cmd = 'lineto'
    print 'stroke'

def mec_test():
    th0, th1 = 0, pi / 4
    fnl = mk_ths_fnl(th0, th1)
    params = solve_mec_nested(th0, th1, fnl)
    k, lam1, lam2 = params
    xys, cost, x, y, th = run_mec_cos(k, lam1, lam2, 1000)
    plot(xys)
    print '% run_mec_cos:', cost, x, y, th
    print '1 0 0 setrgbcolor'
    xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2)
    print '%', xys
    plot(xys)
    print '% run_elastica:', cost, x, y, th
    print 'showpage'
    print '%', params

def lenfig():
    print '306 720 translate'
    th0, th1 = pi / 2, pi / 2
    for i in range(1, 10):
	y = .1 * i + .003
	fnl = mk_y_fnl(th0, th1, y)
	params = solve_mec_3constr(fnl)
	k, lam1, lam2 = params
	print 'gsave 0.5 dup scale -306 -396 translate'
	xys, cost, x, y, th = run_elastica(-2, 2, k, lam1, lam2, 100)
        print 'gsave [2] 0 setdash'
	plot(xys)
        print 'grestore'
	xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2, 100)
	plot(xys)
	print 'grestore'
        print '% y =', y, 'params =', params
        if lam2 < 0:
            mymaxk = k
        else:
            mymaxk = sqrt(k * k + 4 * lam2)
        lam = abs(lam2) / (mymaxk * mymaxk)
        print '-200 0 moveto /Symbol 12 selectfont (l) show'
        print '/Times-Roman 12 selectfont ( = %.4g) show' % lam
	print '0 -70 translate'
    print 'showpage'

def lenplot(figno, th0, th1):
    result = []
    if figno in (1, 2):
	imax = 181
    elif figno == 3:
	imax = 191
    for i in range(1, imax):
        yy = .005 * i
        if figno in (1, 2) and i == 127:
            yy = 2 / pi
	if figno == 3 and i == 165:
	    yy = .8270
	fnl = mk_y_fnl(th0, th1, yy)
	params = solve_mec_3constr(fnl)
	k, lam1, lam2 = params
        xys, cost, x0, y0, th = run_elastica(-.5, .5, k, lam1, lam2, 100)
        if lam2 < 0:
            mymaxk = k
        else:
            mymaxk = sqrt(k * k + 4 * lam2)
        lam = abs(lam2) / (mymaxk * mymaxk)
        result.append((yy, lam, cost))
    return result

def lengraph(figno):
    if figno in (1, 2):
        eps_prologue(15, 47, 585, 543)
	th0, th1 = pi / 2, pi / 2
	yscale = 900
	ytic = .05
	ymax = 10
    elif figno == 3:
        eps_prologue(15, 47, 585, 592)
	th0, th1 = pi / 3, pi / 3
	yscale = 500
	ytic = .1
	ymax = 10
    x0 = 42
    xscale = 7.5 * 72
    y0 = 72
    print '/Times-Roman 12 selectfont'
    print '/ss 1.5 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'
    print '.5 setlinewidth'
    print x0, y0, 'moveto', xscale, '0 rlineto 0', yscale * ytic * ymax, 'rlineto', -xscale, '0 rlineto closepath stroke'
    for i in range(0, 11):
        print x0 + .1 * i * xscale, y0, 'moveto 0 6 rlineto stroke'
        print x0 + .1 * i * xscale, y0 + ytic * ymax * yscale, 'moveto 0 -6 rlineto stroke'
        print x0 + .1 * i * xscale, y0 - 12, 'moveto'
        print '(%.1g) dup stringwidth exch -.5 mul exch rmoveto show' % (.1 * i)
    for i in range(0, 11):
        print x0, y0 + ytic * i * yscale, 'moveto 6 0 rlineto stroke'
        print x0 + xscale, y0 + ytic * i * yscale, 'moveto -6 0 rlineto stroke'
        print x0 - 3, y0 + ytic * i * yscale - 3.5, 'moveto'
        print '(%.2g) dup stringwidth exch neg exch rmoveto show' % (ytic * i)
    print x0 + .25 * xscale, y0 - 24, 'moveto (chordlen / arclen) show'
    print x0 - 14, y0 + ytic * ymax * yscale + 10, 'moveto /Symbol 12 selectfont (l) show'
    if figno in (1, 2):
	print x0 + 2 / pi * xscale, y0 - 18, 'moveto'
	print '(2/p) dup stringwidth exch -.5 mul exch rmoveto show'
    print 'gsave [3] 0 setdash'
    print x0, y0 + .25 * yscale, 'moveto', xscale, '0 rlineto stroke'
    if figno == 3:
	print x0, y0 + .5 * yscale, 'moveto', xscale, '0 rlineto stroke'
	xinterest = .81153
	print x0 + xinterest * xscale, y0, 'moveto 0', yscale * .5, 'rlineto stroke'
    print 'grestore'
    print '.75 setlinewidth'
    if 1:
	if figno in (1, 2):
	    costscale = .01 * yscale
	elif figno == 3:
	    costscale = .0187 * yscale
        lenresult = lenplot(figno, th0, th1)
        cmd = 'moveto'
        for x, y, cost in lenresult:
            print x0 + xscale * x, y0 + yscale * y, cmd
            cmd = 'lineto'
        print 'stroke'
	if figno in (2, 3):
            cmd = 'moveto'
            for x, y, cost in lenresult:
                print x0 + xscale * x, y0 + costscale * cost, cmd
                cmd = 'lineto'
            print 'stroke'
            cmd = 'moveto'
            for x, y, cost in lenresult:
                print x0 + xscale * x, y0 + costscale * x * cost, cmd
                cmd = 'lineto'
            print 'stroke'
            print '/Times-Roman 12 selectfont'
	    if figno == 2:
		xlm, ylm = .75, 7
		xls, yls = .42, 15
	    elif figno == 3:
		xlm, ylm = .6, 3
		xls, yls = .37, 15
            print x0 + xscale * xlm, y0 + costscale * ylm, 'moveto'
            print '(MEC cost) show'
            print x0 + xscale * xls, y0 + costscale * yls, 'moveto'
            print '(SIMEC cost) show'
    if figno == 1:
        minis = [(.05, 5, -5),
             (.26, -40, 10),
             (.4569466, -5, -10),
             (.55, 35, 12),
             
             (.6026, -60, 45),
             (.6046, -60, 15),
             (.6066, -60, -15),
             
             (.6213, -22, 10),
             (.6366198, 15, 22),
             (.66, 20, 10),
             (.9, 0, -10)]
    elif figno == 2:
        minis = [(.4569466, -5, -10),
             (.6366198, 15, 22)]
    elif figno == 3:
	minis = [(.741, 55, -10),
		 (.81153, -30, 20),
		 (.8270, 20, 30)]

    for yy, xo, yo in minis:
	fnl = mk_y_fnl(th0, th1, yy)
	params = solve_mec_3constr(fnl)
	k, lam1, lam2 = params
        if lam2 < 0:
            mymaxk = k
        else:
            mymaxk = sqrt(k * k + 4 * lam2)
        lam = abs(lam2) / (mymaxk * mymaxk)
        x = x0 + xscale * yy
        y = y0 + yscale * lam
	print 'gsave %g %g translate circle fill' % (x, y)
        print '%g %g translate 0.15 dup scale' % (xo, yo)
        print '-306 -396 translate'
        print '2 setlinewidth'
        if yy < .6 or yy > .61:
            s = 2
        elif yy == .6046:
            s = 2.8
        else:
            s = 5
	xys, cost, x, y, th = run_elastica(-s, s, k, lam1, lam2, 100)
        print 'gsave [10] 0 setdash'
	plot(xys)
        print 'grestore 6 setlinewidth'
	xys, cost, x, y, th = run_elastica(-.5, .5, k, lam1, lam2, 100)
	plot(xys)
	print 'grestore'
        
    print 'showpage'
    eps_trailer()

def draw_axes(x0, y0, xscale, yscale, xmax, ymax, nx, ny):
    print '.5 setlinewidth'
    print '/Times-Roman 12 selectfont'
    print x0, y0, 'moveto', xscale * xmax, '0 rlineto 0', yscale * ymax, 'rlineto', -xscale * xmax, '0 rlineto closepath stroke'
    xinc = (xmax * xscale * 1.) / nx
    yinc = (ymax * yscale * 1.) / ny
    for i in range(0, nx + 1):
        if i > 0 and i < nx + 1:
            print x0 + xinc * i, y0, 'moveto 0 6 rlineto stroke'
            print x0 + xinc * i, y0 + ymax * yscale, 'moveto 0 -6 rlineto stroke'
        print x0 + xinc * i, y0 - 12, 'moveto'
        print '(%.2g) dup stringwidth exch -.5 mul exch rmoveto show' % ((i * xmax * 1.) / nx)
    for i in range(0, ny + 1):
        if i > 0 and i < ny + 1:
            print x0, y0 + yinc * i, 'moveto 6 0 rlineto stroke'
            print x0 + xmax * xscale, y0 + yinc * i, 'moveto -6 0 rlineto stroke'
        print x0 - 3, y0 + yinc * i - 3.5, 'moveto'
        print '(%.2g) dup stringwidth exch neg exch rmoveto show' % ((i * ymax * 1.) / ny)
    

def mecchord():
    x0 = 72
    y0 = 72
    thscale = 150
    chscale = 400
    print '.5 setlinewidth'
    print '/ss 1.5 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'
    draw_axes(x0, y0, thscale, chscale, 3.2, 1, 16, 10)
    print x0 + 1 * thscale, y0 - 28, 'moveto (angle) show'
    print 'gsave', x0 - 32, y0 + .25 * chscale, 'translate'
    print '90 rotate 0 0 moveto (chordlen / arclen) show'
    print 'grestore'

    cmd = 'moveto'
    for i in range(0, 67):
        s = i * .01
        xys, cost, x, y, th = run_elastica(-s, s, 4, 0, -8, 100)
        #plot(xys)
        if s > 0:
            ch = hypot(x, y) / (2 * s)
        else:
            ch = 1
        print '%', s, x, y, th, ch
        print x0 + thscale * th / 2, y0 + ch * chscale, cmd
        cmd = 'lineto'
    print 'stroke'

    print 'gsave %g %g translate circle fill' % (x0 + thscale * th / 2, y0 + ch * chscale)
    print '-30 15 translate 0.25 dup scale'
    print '-306 -396 translate'
    print '2 setlinewidth'
    plot(xys)
    print 'grestore'

    print 'gsave [2] 0 setdash'
    cmd = 'moveto'
    for i in range(0, 151):
        th = pi * i / 150.
        if th > 0:
            ch = sin(th) / th
        else:
            ch = 1
        print x0 + thscale * th, y0 + ch * chscale, cmd
        cmd = 'lineto'
    print 'stroke'
    print 'grestore'

    k0 = 4 * .4536 / (2 / pi)
    s = pi / (2 * k0)
    xys, cost, x, y, th = run_elastica(-s, s, k0, 0, 0, 100)
    th = pi
    ch = 2 / pi
    print 'gsave %g %g translate circle fill' % (x0 + thscale * th / 2, y0 + ch * chscale)
    print '30 15 translate 0.25 dup scale'
    print '-306 -396 translate'
    print '2 setlinewidth'
    plot(xys)
    print 'grestore'

    print x0 + 1.25 * thscale, y0 + .55 * chscale, 'moveto (MEC) show'
    print x0 + 2.3 * thscale, y0 + .35 * chscale, 'moveto (SIMEC) show'

    print 'showpage'

def trymec(sm, sp):
    xys, thm, cost = run_elastica_half(abs(sm), 0, 1, 0, 10)
    xm, ym, km = xys[-1]
    if sm < 0:
        xm = -xm
        ym = -ym
        thm = -thm
    xys, thp, cost = run_elastica_half(abs(sp), 0, 1, 0, 10)
    xp, yp, kp = xys[-1]
    if sp < 0:
        xp = -xp
        yp = -yp
        thp = -thp
    chth = atan2(yp - ym, xp - xm)
    #print xm, ym, xp, yp, chth, thm, thp
    actual_th0 = chth - thm
    actual_th1 = thp - chth
    return actual_th0, actual_th1

def findmec_old(th0, th1):
    guess_ds = sqrt(6 * (th1 - th0))
    guess_savg = 2 * (th1 + th0) / guess_ds
    sm = .5 * (guess_savg - guess_ds)
    sp = .5 * (guess_savg + guess_ds)
    #sm, sp = th0, th1
    for i in range(1):
        actual_th0, actual_th1 = trymec(sm, sp)
        guess_dth = 1./6 * (sp - sm) ** 2
        guess_avth = .5 * (sp + sm) * (sp - sm)
        guess_th0 = .5 * (guess_avth - guess_dth)
        guess_th1 = .5 * (guess_avth + guess_dth)
        print sm, sp, actual_th0, actual_th1, guess_th0, guess_th1

def mecplots():
    print '2 dup scale'
    print '-153 -296 translate'
    scale = 45
    print '0.25 setlinewidth'
    print 306 - scale * pi/2, 396, 'moveto', 306 + scale * pi/2, 396, 'lineto stroke'
    print 306, 396 - scale * pi/2, 'moveto', 306, 396 + scale * pi/2, 'lineto stroke'
    print '/ss .5 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'

    tic = .1
    maj = 5
    r0, r1 = 0, 81

    for i in range(r0, r1, maj):
        sm = i * tic
        cmd = 'moveto'
        for j in range(i, r1):
            sp = j * tic + 1e-3
            th0, th1 = trymec(sm, sp)
            print '%', sm, sp, th0, th1
            print 306 + scale * th0, 396 + scale * th1, cmd
            cmd = 'lineto'
        print 'stroke'
    for j in range(r0, r1, maj):
        sp = j * tic + 1e-3
        cmd = 'moveto'
        for i in range(0, j + 1):
            sm = i * tic
            th0, th1 = trymec(sm, sp)
            print '%', sm, sp, th0, th1
            print 306 + scale * th0, 396 + scale * th1, cmd
            cmd = 'lineto'
        print 'stroke'

    for i in range(r0, r1, maj):
        sm = i * tic
        for j in range(i, r1, maj):
            sp = j * tic + 1e-3
            th0, th1 = trymec(sm, sp)
            print 'gsave'
            print 306 + scale * th0, 596 + scale * th1, 'translate'
            print 'circle fill'
            uscale = 3
            xys, thm, cost = run_elastica_half(abs(sm), 0, 1, 0, 100)
            x0, y0 = xys[-1][0], xys[-1][1]
            if sm < 0:
                x0, y0 = -x0, -y0
            xys, thm, cost = run_elastica_half(abs(sp), 0, 1, 0, 100)
            x1, y1 = xys[-1][0], xys[-1][1]
            if sp < 0:
                x1, y1 = -x1, -y1
            cmd = 'moveto'
            for xy in xys:
                print xy[0] * uscale, xy[1] * uscale, cmd
                cmd = 'lineto'
            print 'stroke'
            print '1 0 0 setrgbcolor'
            print x0 * uscale, y0 * uscale, 'moveto'
            print x1 * uscale, y1 * uscale, 'lineto stroke'
            print 'grestore'

    print 'showpage'

def findmec(th0, th1):
    delta = 1e-3
    if th0 < 0 and th0 - th1 < .1:
        sm = 5.
        sp = 6.
    else:
        sm = 3.
        sp = 5.
    params = [sm, sp]
    lasterr = 1e6
    for i in range(25):
        sm, sp = params
        ath0, ath1 = trymec(sm, sp)
        c1c, c2c = th0 - ath0, th1 - ath1

        err = c1c * c1c + c2c * c2c
        if 0:
            print '%findmec', sm, sp, ath0, ath1, err
            sys.stdout.flush()

        if err < 1e-9:
            return params
        if err > lasterr:
            return None
        lasterr = err


        dc1s = []
        dc2s = []
        for j in range(len(params)):
            params1 = N.array(params)
            params1[j] += delta
            sm, sp = params1
            ath0, ath1 = trymec(sm, sp)
            c1p, c2p = th0 - ath0, th1 - ath1

            params1 = N.array(params)
            params1[j] -= delta
            sm, sp = params1
            ath0, ath1 = trymec(sm, sp)
            c1m, c2m = th0 - ath0, th1 - ath1

            dc1s.append((c1p - c1m) / (2 * delta))
            dc2s.append((c2p - c2m) / (2 * delta))

        jm = N.array([dc1s, dc2s])
        ji = la.inverse(jm)
        dp = N.dot(ji, [c1c, c2c])

        if i < 4:
            scale = .5
        else:
            scale = 1
        params -= scale * dp
        if params[0] < 0: params[0] = 0.
    return params

def mecrange(figtype):
    scale = 130
    eps_prologue(50, 110, 570, 630)
    print -50, 0, 'translate'
    print '0.5 setlinewidth'
    thlmin, thlmax = -pi/2, 2.4
    thrmin, thrmax = -2.2, pi / 2 + .2
    print 306 + scale * thlmin, 396, 'moveto', 306 + scale * thlmax, 396, 'lineto stroke'
    print 306, 396 + scale * thrmin, 'moveto', 306, 396 + scale * thrmax, 'lineto stroke'

    print 'gsave [2] 0 setdash'
    print 306, 396 + scale * pi / 2, 'moveto'
    print 306 + scale * thlmax, 396 + scale * pi / 2, 'lineto stroke'
    print 306 + scale * thlmin, 396 - scale * pi / 2, 'moveto'
    print 306 + scale * thlmax, 396 - scale * pi / 2, 'lineto stroke'
    print 306 + scale * pi / 2, 396 + scale * thrmin, 'moveto'
    print 306 + scale * pi / 2, 396 + scale * thrmax, 'lineto stroke'
    print 'grestore'

    print 306 + 3, 396 + scale * thrmax - 10, 'moveto'
    print '/Symbol 12 selectfont (q) show'
    print 0, -2, 'rmoveto'
    print '/Times-Italic 9 selectfont (right) show'

    print 306 - 18, 396 + scale * pi / 2 - 4, 'moveto'
    print '/Symbol 12 selectfont (p/2) show'
    print 306 + scale * 2.2, 396 - scale * pi / 2 + 2, 'moveto'
    print '/Symbol 12 selectfont (-p/2) show'

    print 306 + scale * pi/2 + 2, 396 + scale * thrmax - 10, 'moveto'
    print '/Symbol 12 selectfont (p/2) show'

    print 306 + scale * 2.2, 396 + 6, 'moveto'
    print '/Symbol 12 selectfont (q) show'
    print 0, -2, 'rmoveto'
    print '/Times-Italic 9 selectfont (left) show'

    print '/ss 0.8 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'
    cmd = 'moveto'
    for i in range(0, 201):
        th = (i * .005 - .75 )* pi
        rmin = 1.5
        rmax = 2.5
        for j in range(20):
            r = (rmin + rmax) * .5
            th0 = r * cos(th)
            th1 = r * sin(th)
            if findmec(th0, th1) == None:
                rmax = r
            else:
                rmin = r
        r = (rmin + rmax) * .5
        th0 = r * cos(th)
        th1 = r * sin(th)
        print '%', r, th, th0, th1
        print 306 + scale * th0, 396 + scale * th1, cmd
        cmd = 'lineto'
        sys.stdout.flush()
    print 'stroke'
    sys.stdout.flush()
        
    for i in range(-11, 12):
        for j in range(-11, i + 1):
            th0, th1 = i * .196, j * .196
            print '%', th0, th1
            params = findmec(th0, th1)
            if params != None:
                sm, sp = params
                print 'gsave'
                print 306 + scale * th0, 396 + scale * th1, 'translate'
                uscale = 22
                k0, lam1, lam2 = justify_mec(sm, sp)
                xys, cost, x, y, th = run_elastica(-.5, .5, k0, lam1, lam2)
                cmdm = 'moveto'
                dx = xys[-1][0] - xys[0][0]
                dy = xys[-1][1] - xys[0][1]
                ch = hypot(dx, dy)
                chth = atan2(dy, dx)
                if figtype == 'mecrange':
                    print 'circle fill'
                    s = uscale * sin(chth) / ch
                    c = uscale * cos(chth) / ch
                    h = -xys[0][0] * s + xys[0][1] * c
                    for xy in xys:
                        print xy[0] * c + xy[1] * s, h + xy[0] * s - xy[1] * c, cmdm
                        cmdm = 'lineto'
                elif figtype == 'mecrangek':
                    ds = 1. / (len(xys) - 1)
                    sscale = 13. / ch
                    kscale = 3 * ch
                    print 'gsave .25 setlinewidth'
                    print sscale * -.5, 0, 'moveto', sscale, 0, 'rlineto stroke'
                    print 'grestore'
                    for l in range(len(xys)):
                        print sscale * (ds * l - 0.5), kscale * xys[l][2], cmdm
                        cmdm = 'lineto'
                print 'stroke'
                print 'grestore'
            sys.stdout.flush()
    print 'showpage'
    eps_trailer()

# given sm, sp in [0, 1, 0] mec, return params for -0.5..0.5 segment
def justify_mec(sm, sp):
    sc = (sm + sp) * 0.5
    xys, thc, cost = run_elastica_half(sc, 0, 1, 0, 100)
    print '% sc =', sc, 'thc =', thc
    k0, lam1, lam2 = xys[-1][2], cos(thc), -sin(thc)
    ds = sp - sm
    k0 *= ds
    lam1 *= ds * ds
    lam2 *= ds * ds
    return [k0, lam1, lam2]

import sys
figname = sys.argv[1]
if len(figname) > 4 and figname[-4:] == '.pdf': figname = figname[:-4]
if figname in ('lengraph1', 'lengraph2', 'lengraph3'):
    lengraph(int(figname[-1]))
    #mec_test()
    #lenfig()
elif figname == 'mecchord':
    mecchord()
elif figname in ('mecrange', 'mecrangek'):
    mecrange(figname)
elif figname == 'mecplots':
    mecplots()
elif figname == 'findmec':
    findmec(-.1, -.1)
    findmec(-.2, -.2)





import math

from . import design, make, shape, profile

from nesoni import config

def grow(loop,thickness):
    extent = loop.extent()
    xsize = extent.xmax-extent.xmin
    ysize = extent.ymax-extent.ymin
    return loop.scale2(
        (xsize+thickness*2.0)/xsize,
        (ysize+thickness*2.0)/ysize
        )


@config.help(
    '[Experimental, not yet working] Make a printable reed.'
    )
class Make_reed(make.Make):
    def run(self):
        diameter = 8.0
        flare = 2.0
        length = 30.0
        
        stem_length = 5.0
        stem_diameter = diameter + 6.0 
        
        thickness0 = 1.0
        thickness1 = 0.05
                
        bottom = shape.circle(diameter)
        top = shape.lens(0.99).with_circumpherence(math.pi*diameter*flare)
        
        stem = shape.circle(stem_diameter)

        reed = shape.extrusion(
            [0.0,stem_length,stem_length,stem_length+length],
            [stem,stem,grow(bottom,thickness0),grow(top,thickness1)]
            )
        inside = shape.extrusion(
            [0.0,stem_length,stem_length+length],
            [bottom,bottom,top]
            )
        
        reed.remove(inside)
        
        #wedge_thickness = 0.2
        #wedge_loop = shape.Loop([
        #    (length*0.5, 0.0),
        #    (length*1.01, wedge_thickness*0.5),
        #    (length*1.01, -wedge_thickness*0.5)
        #    ])
        #wedge = shape.extrusion([-diameter*4,diameter*4],[wedge_loop,wedge_loop])
        #wedge.rotate(0,1,0,-90)
        #reed.remove(wedge)
        
        self.save(reed, 'reed')
        
        



import math

from nesoni import config

from . import shape, profile, make

"""

A typical trumpet mouthpiece (Kelly 5C):

cup
16mm   inner diameter
26.5mm outer diameter
depth of about 9mm

3.75mm entrance to bore
flares over ~75mm to 8mm



Cup diam 11 depth 4 bore 5 almost works with 4mm tenor shawm.

"""


def make_widget(
    tube_length,
    tube_diameter,
    cup_length,
    final_diameter,
    french=False,
):
    a = 0.0
    b = a + tube_length
    c = b + cup_length
    
    initial_outer_diameter = final_diameter
    final_outer_diameter = final_diameter + 14.0
    
    outer_profile = profile.make_profile([
         (a, initial_outer_diameter),
         (c, final_outer_diameter),
    ])
    
    inner_spec = [
         (a, tube_diameter),
    ]
    n = 1 if french else 20
    for i in range(n):
        angle = i*math.pi/2/n
        inner_spec.append((
             b+cup_length*(1-math.cos(angle)),
             tube_diameter+(final_diameter-tube_diameter)*math.sin(angle)
        ))
    inner_spec.append((c,final_diameter))
    inner_profile = profile.make_profile(inner_spec).clipped(-10.0, c+10.0)
    
    bevel = profile.make_profile([
         (c-1.0, 0.0),
         (c, 2.0)
         ])
    outer_profile = outer_profile - bevel
    inner_profile = inner_profile + bevel
    
    outer = shape.extrude_profile(outer_profile)
    inner = shape.extrude_profile(inner_profile)
    widget = outer.copy()
    widget.remove(inner)
    
    return outer, inner, widget


def make_horn(straight_length, length, d1, d2, power):
    k = (d1/d2)**(1.0/power)
    a = k*length/(1.0-k)
    print a, k
    b = d1/(abs(a)**power)
    f = lambda x: b*(abs(x+a)**power)
    
    print f(0.0), f(length)
    
    outer_profile = profile.make_profile([
         (0.0, d2+1.5),
         (straight_length+length-0.5, d2+1.5),
         (straight_length+length, d2+0.5),
    ])
    
    steps = 16
    inner_profile = profile.make_profile(
         [ (0.0, d1) ] + [
         (straight_length+i*length/steps, f(i*length/steps))
         for i in xrange(steps)
         ] + [ (straight_length+length, d2) ])
    
    outer = shape.extrude_profile(outer_profile)
    inner = shape.extrude_profile(inner_profile)
    widget = outer.copy()
    widget.remove(inner)
    
    return outer, inner, widget


@config.help("""\
[Experimental] Make a cornett mouthpiece for a shawm.
""","""\
Glue on or attach using a short length of drinking straw.
""")
@config.Float_flag('bore','Bore diameter mouthpiece connects to.')
class Make_mouthpiece(make.Make, make.Miller):
    bore = 5.0

    def run(self):
        #outer, inner, widget = make_widget(
        #     tube_length = 8.0,
        #     tube_diameter = self.bore,
        #     cup_length = 4.0, #5.0 #6.0,
        #     final_diameter = 9.0, #12.0 #15.0,
        #     )
        outer,inner,widget = make_horn(
            4.0, 8.0,
            self.bore,
            20.0,
            -1.0
            )
        self.save(widget, 'mouthpiece')
        
        #Ripped a strand of wood removing
        #mill = self.miller(widget, [0.0],[0.0])
        
        mill = self.miller(widget, [-5.0,5.0],[])
        self.save(mill, 'mill-mouthpiece')




"""

Units:

distance:  mm
time:      seconds
speed:     mm/s
frequency: Hz

Positions are measured from the end of the instrument,
the mouthpiece is at instrument.length

Fingerholes are enumerated from bottom (end-most) to top.

"""

import math, os, pickle, sys, random

import profile, svg

from nesoni import config

# ===== Utility constants and functions =====

SPEED_OF_SOUND = 346100.0 # mm/sec at 25C
                          #           30C is 349000

semitone_name = [ 'C', 'C#', 'D', 'Eb', 'E', 'F', 'F#', 'G', 'G#', 'A', 'Bb', 'B' ] 

semitone = {
   'C'  : 0,
   'D'  : 2,
   'E'  : 4,
   'F'  : 5,
   'G'  : 7,
   'A'  : 9,
   'B'  : 11,
}


def fqc(note):
    if '*' in note:
        note, mult = note.split('*')
        mult = float(mult)
    else:
        mult = 1.0

    s = semitone[note[0].upper()]
    note = note[1:]
    if note.startswith('b'):
        s -= 1
        note = note[1:]
    if note.startswith('#'):
        s += 1
        note = note[1:]
    
    s += 12 * int(note)
    
    return 440.0 * 2**((s-57)/12.0) * mult


def wavelength(w, transpose=0):
    if isinstance(w, str): 
        w = SPEED_OF_SOUND / fqc(w)
    return w / (2**(transpose/12.0))


def log2(x):
    return math.log(x) / math.log(2.0)


def describe(w):
    f = SPEED_OF_SOUND / w
    
    s = int(round( log2(f/440.0) * 12.0 + 57 ))
    octave = s // 12
    s = s % 12
    
    return semitone_name[s] + str(octave)



# ===== Frequency response of a tree of connected tubes ====

FOUR_PI = 4.0*math.pi

def circle_area(diameter):
    radius = diameter*0.5
    return math.pi * radius * radius

def pipe_reply(reply_end, length_on_wavelength):
    angle = FOUR_PI*length_on_wavelength
    return complex(math.cos(angle),math.sin(angle)) * reply_end


def junction_reply(area, areas, replies):
    total_area = area + sum(areas)
    
    return area / (total_area*(0.5 - sum( a/(total_area*(1.0/r+1.0)) for a,r in zip(areas, replies) ))) - 1.0
    
    #Equivalently?
    #x = area / (total_area*(0.5 - sum( a/(total_area*(r+1.0)) for a,r in zip(areas, replies) ))) - 1.0
    #return 1/x


def junction2_reply(area, area1, reply1):
    total_area = area + area1
    
    return area / (total_area*(0.5 - area1/(total_area*(1.0/reply1+1.0)))) - 1.0


def junction3_reply(area, area1, area2, reply1, reply2):
    total_area = area + area1 + area2
    
    return area / (total_area*(
        0.5 
        - area1/(total_area*(1.0/reply1+1.0))
        - area2/(total_area*(1.0/reply2+1.0))
    )) - 1.0


def end_flange_length_correction(outer_diameter, inner_diameter):
    a = inner_diameter / 2.0
    w = (outer_diameter - inner_diameter) / 2.0
    
    return a * (0.821 - 0.13*((0.42+w/a)**-0.54))


def hole_length_correction(hole_diameter, bore_diameter, closed):
    # No inner correction even, for closed holes.
    # Not sure why, but including an inner correction
    # for closed holes is wrong (I've tried this).
    
    # Maybe better to treat as bore deviation?
    
    if closed: return 0.0
    
    # As per p.63-64 of Nederveen
    outer_correction = 0.7
    inner_correction = 1.3 - 0.9 * hole_diameter/bore_diameter
    a = hole_diameter / 2.0
    
    result = a * (inner_correction+outer_correction)
    return result



# ================ Instruments and classes of instruments =================

class Instrument:
    """
        Fill in:
            length - length
            inner  - inner profile, diameters
            outer  - outer profile, diameters
            hole_positions - hole positions on outside
            hole_angles - angle of hole in degrees, up positive, down negative
            inner_hole_positions - hole positions in bore
            hole_lengths - length of hole through instrument wall (perhaps plus an embouchure correction)
            hole_diameters - hole diameters
            closed_top - is the mouthpiece closed (eg reed) or open (eg ney)
            cone_step - inner profile step size for conical segments of inner profile
        
        Then call:
            .prepare()
               - prepare to handle queries
    
            .evaluate(wavelength)
    """
    
    def prepare(self):
        self.stepped_inner = self.inner.as_stepped(self.cone_step)
    
        events = [
            (self.length, 'end', None)
        ]
        
        for i, pos in enumerate(self.stepped_inner.pos):
            if 0.0 < pos < self.length:
                events.append((pos, 'step', i))
        
        for i, pos in enumerate(self.inner_hole_positions):
            events.append((pos,'hole', i))
        
        events.sort(key=lambda item: item[0])
    
        self.actions = [ ]
        
        position = -end_flange_length_correction(self.outer(0.0,True),self.stepped_inner(0.0,True))
        diameter = self.stepped_inner(0.0, True)
        for pos, action, index in events:
            length = pos-position
            
            def func(reply_end, wavelength, fingers, length=length):
                return pipe_reply(reply_end, length/wavelength)
            self.actions.append(func)
            
            position = pos
            
            if action == 'step':
                assert diameter == self.stepped_inner.low[index]
                area1 = circle_area(diameter)
                diameter = self.stepped_inner.high[index]
                area = circle_area(diameter)
                
                def func(reply, wavelength, fingers, area=area, area1=area1):
                    return junction2_reply(area, area1, reply)
                self.actions.append(func)
                
            elif action == 'hole':
                area = circle_area(diameter)
                hole_diameter = self.hole_diameters[index]
                hole_area = circle_area(hole_diameter)
                
                #true_length = (self.outer(position) - diameter) * 0.5 
                #true_length += self.hole_extra_heights[index]
                #true_length += self.hole_extra_height_by_diameter[index] * hole_diameter                
                true_length = self.hole_lengths[index]
                
                open_length = true_length + hole_length_correction(hole_diameter, diameter, False) 
                closed_length = true_length + hole_length_correction(hole_diameter, diameter, True)

                def func(reply, wavelength, fingers, 
                         area=area,hole_area=hole_area,
                         open_length=open_length,closed_length=closed_length,
                         index=index):
                    if fingers[index]:
                        hole_reply = pipe_reply(1.0, closed_length/wavelength)
                    else:
                        hole_reply = pipe_reply(-1.0, open_length/wavelength)
                    return junction3_reply(area, area, hole_area, reply, hole_reply)
                self.actions.append(func)
    
    
    def resonance_score(self, w, fingers):
        """ A score -1 <= score <= 1, zero if wavelength w resonantes """
        reply = -1.0 #Open end
        for action in self.actions:
            reply = action(reply, w, fingers)
        
        if not self.closed_top: reply *= -1.0
        
        angle = math.atan2(reply.imag, reply.real)
        
        angle1 = angle  %(math.pi*2.0)
        angle2 = angle1 - math.pi*2.0
        
        if angle1 < -angle2:
            return angle1/math.pi
        else:
            return angle2/math.pi
        
        
    def true_wavelength_near(self, w, fingers, max_grad, step_cents = 1.0, step_increase = 1.05, max_steps = 100):
        step = pow(2.0, step_cents/1200.0)
        
        half_step = math.sqrt(step)
        probes = [ w/half_step, w*half_step ]
        scores = [ self.resonance_score(probe,fingers) for probe in probes ]
        
        def evaluate(i):
            y1 = scores[i]
            x1 = probes[i]
            y2 = scores[i+1]
            x2 = probes[i+1]
            
            m = (y2-y1)/(x2-x1)
            c = y1-m*x1
            intercept = -c/m
            
            grad = -m*intercept
            if grad > max_grad: return None
            
            assert x1 <= intercept <= x2
            return intercept, grad
        
        for iteration in range(max_steps):
            if scores[-2] >= 0 and scores[-1] < 0:
                result = evaluate(len(scores)-2)
                if result is not None: return result
            
            probes.insert(0, probes[0]/step)
            scores.insert(0, self.resonance_score(probes[0],fingers))
            
            if scores[0] >= 0 and scores[1] < 0:
                result = evaluate(0)
                if result is not None: return result

            probes.append(probes[-1]*step)
            scores.append(self.resonance_score(probes[-1],fingers))
        
            step **= step_increase
        else:
            if abs(scores[-1]) < abs(scores[0]):
                return probes[-1], 0.0
            else:
                return probes[0], 0.0
        
        # y = mx + c
                
        #step = pow(2.0, max_cents/((n_probes-1)*0.5*1200.0))
        #low = w * pow(step, -(n_probes-1)/2.0)
        #probes = [ low * pow(step,i) for i in range(n_probes) ]
        #
        #scores = [ abs(self.resonance_score(probe, fingers)) for probe in probes ]
        #
        #best = min(range(n_probes), key=lambda i: scores[i])
        #
        #if best == 0 or best == n_probes-1:
        #    return probes[best]
        #    
        #c = scores[best]
        #b = 0.5*(scores[best+1]-scores[best-1])
        #a = scores[best+1]-c-b
        #return low*pow(step, best-b*0.5/a)


def scaler(value):
    @property
    def func(self):
        scale = self.scale
        return [ item*scale if item is not None else None for item in value ]
    return func


def sqrt_scaler(value):
    @property
    def func(self):
        scale = math.sqrt(self.scale)
        return [ item*scale if item is not None else None for item in value ]
    return func


@config.Int_flag('transpose', 'Transpose instrument by this many semitones.')
class Instrument_designer(config.Action_with_output_dir):
    instrument_class = Instrument

    cone_step = 0.125  #diameter step size when approximating cones

    closed_top = False

    transpose = 0
    
    initial_length = None

    fingerings = [ ]
    
    inner_diameters = [ ]             #diameters of inner bore: first is diam at bottom, last is diam at top
    @property
    def inner_angles(self): return [ None ] * len(self.inner_diameters)
    
    outer_add = False
    
    outer_diameters = [ ]
    @property
    def outer_angles(self): return [ None ] * len(self.outer_diameters)
    
    max_hole_diameters = [ ]

    top_clearance_fraction = 0.0
    bottom_clearance_fraction = 0.0

    @property
    def scale(self):
        return 2.0**(-self.transpose/12.0)

    @property
    def n_holes(self):
        return len(self.max_hole_diameters)

    @property
    def min_hole_diameters(self):
        return [ 0.5 ] * self.n_holes
    
    @property
    def min_hole_spacing(self):
        return [ 0.0 ] * (self.n_holes-1)

    @property
    def max_hole_spacing(self):
        return [ self.initial_length ] * (self.n_holes-1)

    @property
    def balance(self):
        return [ None ] * (self.n_holes-2)
    
    # Angle in degrees up (positive) or down (negative) the tube
    @property
    def hole_angles(self):
        return [ 0.0 ] * self.n_holes

    #n-2 initial positions of bore kinks, as fraction of initial length
    @property
    def initial_inner_fractions(self):
        return [ (i+1.0)/(len(self.inner_diameters)-1) for i in xrange(len(self.inner_diameters)-2) ]

    @property
    def min_inner_fraction_sep(self):
        return [ 0.0 ] * (len(self.inner_diameters)-1)

    @property
    def initial_outer_fractions(self):
        return [ (i+1.0)/(len(self.outer_diameters)-1) for i in xrange(len(self.outer_diameters)-2) ]

    @property
    def min_outer_fraction_sep(self):
        return [ 0.0 ] * (len(self.outer_diameters)-1)

    @property
    def initial_hole_fractions(self):
        return [ (i+3.0)/(self.n_holes+2) * 0.5 for i in xrange(self.n_holes) ]

    @property
    def initial_hole_diameter_fractions(self):
        return [ 0.75 ] * self.n_holes

    @property
    def hole_extra_height_by_diameter(self):
        return [ 0.0 ] * self.n_holes

    @property
    def initial_state_vec(self):
        assert len(self.initial_hole_fractions) == self.n_holes, 'initial_hole_fractions has wrong length'
        assert len(self.initial_hole_diameter_fractions) == self.n_holes, 'initial_hole_diameter_fractions has wrong length'
        assert len(self.initial_inner_fractions) == len(self.inner_diameters)-2, 'initial_inner_fractions has wrong length'
        assert len(self.initial_outer_fractions) == len(self.outer_diameters)-2, 'initial_outer_fractions has wrong length'
        return (
            [ 1.0 ] +
            self.initial_hole_fractions + 
            [ item*item for item in self.initial_hole_diameter_fractions ] +
            self.initial_inner_fractions +
            self.initial_outer_fractions
        )

    #
    #Sometimes there are tiny glitches in the response curve
    #that aren't really playable or desirable solutions
    #
    #Set this small enough to forbid them (see the grad numbers in the .SVG output)
    #
    max_grad = 1e30

    def unpack(self, state_vec):
        inst = self.instrument_class()
        
        scale = self.scale
        
        p = 0
        inst.length = state_vec[0] * self.initial_length * scale
        p += 1        
        inst.hole_positions = [ item*inst.length
                                for item in state_vec[p:p+self.n_holes] ]
        p += self.n_holes
        signed_sqrt = lambda x: math.sqrt(abs(x))*(1 if x >= 0 else -1)
        inst.hole_diameters = [ self.min_hole_diameters[i]+signed_sqrt(item)*(self.max_hole_diameters[i]-self.min_hole_diameters[i])
                                for i, item in enumerate(state_vec[p:p+self.n_holes]) ]
        p += self.n_holes
        
        inner_kinks = [ item*inst.length 
                        for item in state_vec[p:p+len(self.inner_diameters)-2] ]
        p += len(self.inner_diameters)-2
        outer_kinks = [ item*inst.length
                        for item in state_vec[p:p+len(self.outer_diameters)-2] ]
        p += len(self.outer_diameters)-2
        assert p == len(state_vec)
        
        inst.inner = profile.curved_profile(
            [0.0]+inner_kinks+[inst.length],
            self.inner_diameters,
            self.inner_diameters,
            self.inner_angles,
            self.inner_angles,
        )
        inst.outer = profile.curved_profile(
            [0.0]+outer_kinks+[inst.length],
            self.outer_diameters,
            self.outer_diameters,
            self.outer_angles,
            self.outer_angles,
        )
        
        if self.outer_add:
            inst.outer = inst.outer + inst.inner

        inst.hole_angles = self.hole_angles        
        inst.inner_hole_positions = [ None ] * self.n_holes
        inst.hole_lengths = [ None ] * self.n_holes
        for i in range(self.n_holes):
            #Note: approximates bore as cylindrical calculating shift, length
            radians = inst.hole_angles[i]*math.pi/180.0
            thickness = (inst.outer(inst.hole_positions[i])-inst.inner(inst.hole_positions[i])) * 0.5
            shift = math.sin(radians) * thickness
            inst.inner_hole_positions[i] = inst.hole_positions[i] + shift
            inst.hole_lengths[i] = (
                math.sqrt(thickness*thickness+shift*shift) +
                self.hole_extra_height_by_diameter[i] * inst.hole_diameters[i]
            )
        
        inst.inner_kinks = inner_kinks
        inst.outer_kinks = outer_kinks
        
        inst.cone_step = self.cone_step
        inst.closed_top = self.closed_top
        return inst

    
    def constraint_score(self, inst):
        """ Return an amount of constraint dissatisfaction """
        scores = [ ]
        
        #if inst.length < 0: 
            #score += -inst.length
        scores.append(inst.length)
        
        inners = [0.0]+inst.inner_kinks+[inst.length]
        for i in range(len(inners)-1):
            check = inners[i+1] - inners[i] - self.min_inner_fraction_sep[i]*inst.length
            #if check < 0: score += -check 
            scores.append(check)
        
        outers = [0.0]+inst.outer_kinks+[inst.length]
        for i in range(len(outers)-1):
            check = outers[i+1] - outers[i] - self.min_outer_fraction_sep[i]*inst.length
            #if check < 0: score += -check
            scores.append(check)
        
        if self.n_holes:
            check = inst.hole_positions[0] - self.bottom_clearance_fraction*inst.length
            #if check < 0: score += -check
            scores.append(check)
            
            check = (1.0-self.top_clearance_fraction)*inst.length - inst.hole_positions[-1]
            #if check < 0: score += -check
            scores.append(check)
            
            for i, value in enumerate(self.min_hole_spacing):
                if value is None: continue
                check = (inst.hole_positions[i+1]-inst.hole_positions[i]) - value
                #if check < 0: score += -check
                scores.append(check)
            
            for i, value in enumerate(self.max_hole_spacing):
                if value is None: continue
                check = value - (inst.hole_positions[i+1]-inst.hole_positions[i])
                #if check < 0: score += -check
                scores.append(check)                
            
            for i, value in enumerate(self.min_hole_diameters):
                check = inst.hole_diameters[i] - value
                #if check < 0: score += -check
                scores.append(check)
            
            for i, value in enumerate(self.max_hole_diameters):
                if value is None: continue
                check = value - inst.hole_diameters[i]
                #if check < 0: score += -check
                scores.append(check)
            
            for i, value in enumerate(self.balance):
                if value is None: continue
                check = value * 0.5*(inst.hole_positions[i+2]-inst.hole_positions[i]) - abs(
                    0.5*inst.hole_positions[i]+0.5*inst.hole_positions[i+2]-inst.hole_positions[i+1]
                )
                #if check < 0: score += -check
                scores.append(check)

        negscores = [ -item for item in scores if item < 0.0 ]
        negscore = sum(negscores) if negscores else 0.0
        #posscores = [ (item+1e-30)**-1 for item in scores if item >= 0.0 ]
        #posscore = 1.0/sum(posscores) if posscores else 1e30
        return negscore

    def patch_instrument(self, inst):
        """ Hook to modify instrument before scoring. """
        return inst
    
    def score(self, inst):
        inst = self.patch_instrument(inst)
    
        score = 0.0
        div = 0.0
        
        inst.prepare()
        
        s = 1200.0/math.log(2)
        for note, fingers in self.fingerings:
            w1 = wavelength(note, self.transpose)
            w2, grad = inst.true_wavelength_near(w1, fingers, self.max_grad)
            diff = abs(math.log(w1)-math.log(w2))*s
            #weight = w1
            weight = 1.0
            score += weight * diff**3 / (1.0 + (diff/20.0)**2)
            div += weight
            
        return (score/div)**(1.0/3)
        
        #return ( (score/scale) ** (1.0/2) )*100.0

    def _constrainer(self, state_vec):
        return self.constraint_score(self.unpack(state_vec))

    def _scorer(self, state_vec):
        return self.score(self.unpack(state_vec))

    #def _opt_score(self, state_vec):
    #    inst = self.unpack(state_vec)
    #    
    #    cs = self.constraint_score(inst)
    #    if cs:
    #        result = (cs, 0.0)
    #    else:
    #        result = (0.0, self.score(inst))
    #    return result

    
    def _draw(self, diagram, state_vec, color='#000000', red_color='#ff0000'):
        instrument = self.unpack( state_vec )
        instrument.prepare()
        
        for i in range(self.n_holes):
            diagram.circle(0.0, -instrument.inner_hole_positions[i], instrument.hole_diameters[i],
                           red_color)
            diagram.circle(0.0, -instrument.hole_positions[i], instrument.hole_diameters[i], color)
        diagram.profile(instrument.outer, color)
        diagram.profile(instrument.stepped_inner, color)
        
        if self.closed_top:
            d = instrument.stepped_inner(instrument.length)
            diagram.line([(-0.5*d,-instrument.length),(0.5*d,-instrument.length)], color)
        
        tick_x = instrument.outer.maximum() * -0.625
        for pos in instrument.inner_kinks:
            diagram.line([(tick_x,-pos),(tick_x-5,-pos)], color)
        for pos in instrument.inner.pos[1:-1]:
            diagram.line([(tick_x-2,-pos),(tick_x-3,-pos)], color)
        for pos in instrument.outer_kinks:
            diagram.line([(tick_x-10,-pos),(tick_x-15,-pos)], color)
        for pos in instrument.outer.pos[1:-1]:
            diagram.line([(tick_x-12,-pos),(tick_x-13,-pos)], color)

    
    def _save(self, state_vec, other_vecs=[]):
        self.instrument = self.unpack( state_vec )
        
        with open(os.path.join(self.output_dir, 'data.pickle'), 'wb') as f:
            if sys.version_info.major == 3:
                pickle.dump(self, f, fix_imports=True)
            else:
                pickle.dump(self, f)

        patched_instrument = self.patch_instrument(self.instrument)
        patched_instrument.prepare()
        
        #any_extra = any( item != 0.0 for item in self.hole_extra_height_by_diameter )
        #if any_extra:
        #TODO: implement embextra using patch_instrument
        mod_instrument = self.unpack( state_vec )
        #mod_instrument.hole_extra_height_by_diameter = [ 0.0 ] * len(self.hole_extra_height_by_diameter)
        mod_instrument.hole_lengths = [
            (mod_instrument.outer(mod_instrument.hole_positions[i])
             - mod_instrument.inner(mod_instrument.hole_positions[i])) * 0.5
            for i in range(self.n_holes)
        ]
        mod_instrument.prepare()
        
        diagram = svg.SVG()
        
        for vec in random.sample(other_vecs, min(20,len(other_vecs))):
            self._draw(diagram, vec, '#aaaaaa', '#ffaaaa')
        
        self._draw(diagram, state_vec)
        
        #for i in range(self.n_holes):
        #    diagram.circle(0.0, -self.instrument.inner_hole_positions[i], self.instrument.hole_diameters[i],
        #                   '#ff0000')
        #    diagram.circle(0.0, -self.instrument.hole_positions[i], self.instrument.hole_diameters[i])
        #diagram.profile(self.instrument.outer)
        #diagram.profile(self.instrument.stepped_inner)
        #
        #if self.closed_top:
        #    d = self.instrument.stepped_inner(self.instrument.length)
        #    diagram.line([(-0.5*d,-self.instrument.length),(0.5*d,-self.instrument.length)])
            
        text_x = diagram.max_x * 1.25
        text_y = 0
        for i in range(self.n_holes):
            this_y = min(text_y, -self.instrument.hole_positions[i])
            text_y = diagram.text(text_x, this_y, '%.1fmm' % self.instrument.hole_diameters[i])
            diagram.text(text_x + 90.0, this_y, 'at %.1fmm' % self.instrument.hole_positions[i])
            
            if i < self.n_holes-1:
                this_y = min(text_y, -0.5*(self.instrument.hole_positions[i]+self.instrument.hole_positions[i+1]))
                #text_y = 
                diagram.text(text_x + 45.0, this_y,
                   '%.1fmm' % (self.instrument.hole_positions[i+1]-self.instrument.hole_positions[i])
                )
        
        diagram.text(text_x + 90.0, min(text_y,-self.instrument.length), '%.1fmm' % self.instrument.length)

        text_x = diagram.max_x
        text_y = 0
        for note, fingers in self.fingerings:
            w1 = wavelength(note, self.transpose)
            w2, grad = patched_instrument.true_wavelength_near(w1, fingers, self.max_grad)
            cents = int(round( log2(w2/w1) * 1200.0 ))

            n_probes = 301
            max_cents = 2400.0
            width = 200.0
            step = pow(0.5, max_cents/((n_probes-1)*0.5*1200.0))
            low = w1 * pow(step, -(n_probes-1)/2.0)
            probes = [ low * pow(step,i) for i in range(n_probes) ]
            scores = [ patched_instrument.resonance_score(probe, fingers) for probe in probes ]

            graph_x = text_x+200
            points = [ (graph_x+i*width/n_probes,text_y-score*7.0)
                       for i,score in enumerate(scores) ]
            for i in range(len(probes)-1):
                if scores[i] > 0 and scores[i+1] < 0: continue
                diagram.line(points[i:i+2], '#000000', 0.2)
            diagram.line([(graph_x+width*0.5,text_y+7),(graph_x+width*0.5,text_y-7)], '#0000ff', 0.2)
            diagram.line([(graph_x,text_y),(graph_x+width,text_y)], '#0000ff', 0.2)

            diagram.text(text_x, text_y,
                  '%5s %s %-4d cents grad=%.1f' % (describe(w1), '     ' if cents == 0 else ' flat' if cents > 0 else 'sharp', abs(cents), grad)
            )
            
            #if any_extra:
            w3, grad3 = mod_instrument.true_wavelength_near(w1, fingers, self.max_grad)
            cents3 = int(round( log2(w3/w1) * 1200.0 ))
            diagram.text(graph_x + width, text_y,
                  '%s %-4d cents grad=%.1f diff=%d' % ('     ' if cents3 == 0 else ' flat' if cents3 > 0 else 'sharp', abs(cents3), grad3, cents3-cents)
            )
            
            
            text_y -= 25
            
            
        
        diagram.text(0.0, 20.0, 'Outer diameters: %s' % ', '.join('%.1fmm'%item for item in self.outer_diameters))
        diagram.text(0.0, 40.0, 'Inner diameters: %s' % ', '.join('%.1fmm'%item for item in self.inner_diameters))
        
        diagram.save( os.path.join(self.output_dir, 'diagram.svg') )
        
        del self.instrument

    def run(self):
        import optimize

        assert self.initial_length is not None, 'Initial length required'
        assert len(self.min_hole_diameters) == self.n_holes
        assert len(self.max_hole_diameters) == self.n_holes
        assert len(self.inner_diameters) >= 2
        assert len(self.outer_diameters) >= 2

        if not os.path.exists(self.output_dir):
            os.mkdir(self.output_dir)

        state_vec = self.initial_state_vec
        state_vec = optimize.improve(self.shell_name(), self._constrainer, self._scorer, state_vec, monitor=self._save)
        
        self._save(state_vec)
        
        #print(self._opt_score(state_vec))


def load(output_dir):
    with open(os.path.join(output_dir, 'data.pickle'), 'rb') as f:
        if sys.version_info.major == 3:
            return pickle.load(f, fix_imports=True)
        else:
            return pickle.load(f)




#=======================
# Main

def main(module_path, class_name):
    # Explicitly import module
    # If class defined in __main__ module, pickles won't load
    module = __import__(module_path, globals(), fromlist=True)
    designer = getattr(module, class_name)()    
    
    args = sys.argv[1:]
    
    config.shell_run(designer, args, sys.executable + ' -m ' + module_path)
    





#!/usr/bin/env python

import demakein, nesoni

class Design_my_whistle(demakein.Design_folk_whistle):
    transpose = 12    
    
    # ... and any other things you want to change ...

if __name__ == '__main__':
    nesoni.run_tool(Design_my_whistle)


from math import *
import array
import sys
import random

def run_mvc(k, k1, k2, k3, C, n = 100, do_print = False):
    cmd = 'moveto'
    result = array.array('d')
    cost = 0
    th = 0
    x = 0
    y = 0
    dt = 1.0 / n
    for i in range(n):
        k4 = -k * (k * k2 - .5 *  k1 * k1 + C)

        cost += dt * k1 * k1
        x += dt * cos(th)
        y += dt * sin(th)
        th += dt * k

        k += dt * k1
        k1 += dt * k2
        k2 += dt * k3
        k3 += dt * k4
        result.append(k)
        if do_print: print 400 + 400 * x, 500 + 400 * y, cmd
        cmd = 'lineto'
    return result, cost, x, y, th

def run_mec_cos(k, lam1, lam2, n = 100, do_print = False):
    cmd = 'moveto'
    result = array.array('d')
    cost = 0
    th = 0
    x = 0
    y = 0
    dt = 1.0 / n
    for i in range(n):
        k1 = lam1 * cos(th) + lam2 * sin(th)

        cost += dt * k * k
        x += dt * cos(th)
        y += dt * sin(th)
        th += dt * k

        k += dt * k1
        result.append(k)
        if do_print: print 400 + 400 * x, 500 + 400 * y, cmd
        cmd = 'lineto'
    return result, cost, x, y, th

def descend(params, fnl):
    delta = 1
    for i in range(100):
        best = fnl(params, i, True)
        bestparams = params
        for j in range(2 * len(params)):
            ix = j / 2
            sign = 1 - 2 * (ix & 1)
            newparams = params[:]
            newparams[ix] += delta * sign
            new = fnl(newparams, i)
            if (new < best):
                bestparams = newparams
                best = new
        if (bestparams == params):
            delta *= .5
        print '%', params, delta
        sys.stdout.flush()
        params = bestparams
    return bestparams

def descend2(params, fnl):
    delta = 20
    for i in range(5):
        best = fnl(params, i, True)
        bestparams = params
        for j in range(100000):
            newparams = params[:]
            for ix in range(len(params)):
                newparams[ix] += delta * (2 * random.random() - 1)
            new = fnl(newparams, i)
            if (new < best):
                bestparams = newparams
                best = new
        if (bestparams == params):
            delta *= .5
        params = bestparams
        print '%', params, best, delta
        sys.stdout.flush()
    return bestparams

def desc_eval(params, dfdp, fnl, i, x):
    newparams = params[:]
    for ix in range(len(params)):
        newparams[ix] += x * dfdp[ix]
    return fnl(newparams, i)

def descend3(params, fnl):
    dp = 1e-6
    for i in range(1000):
        base = fnl(params, i, True)
        dfdp = []
        for ix in range(len(params)):
            newparams = params[:]
            newparams[ix] += dp
            new = fnl(newparams, i)
            dfdp.append((new - base) / dp)
        print '% dfdp = ', dfdp
        xr = 0.
        yr = base
        xm = -1e-3
        ym = desc_eval(params, dfdp, fnl, i, xm)
        if ym > yr:
            while ym > yr:
                xl, yl = xm, ym
                xm = .618034 * xl
                ym = desc_eval(params, dfdp, fnl, i, xm)
        else:
            xl = 1.618034 * xm
            yl = desc_eval(params, dfdp, fnl, i, xl)
            while ym > yl:
                xm, ym = xl, yl
                xl = 1.618034 * xm
                yl = desc_eval(params, dfdp, fnl, i, xl)

        # We have initial bracket; ym < yl and ym < yr

        x0, x3 = xl, xr
        if abs(xr - xm) > abs(xm - xl):
            x1, y1 = xm, ym
            x2 = xm + .381966 * (xr - xm)
            y2 = desc_eval(params, dfdp, fnl, i, x2)
        else:
            x2, y2 = xm, ym
            x1 = xm + .381966 * (xl - xm)
            y1 = desc_eval(params, dfdp, fnl, i, x1)
        for j in range(30):
            if y2 < y1:
                x0, x1, x2 = x1, x2, x2 + .381966 * (x3 - x2)
                y0, y1 = y1, y2
                y2 = desc_eval(params, dfdp, fnl, i, x2)
            else:
                x1, x2, x3 = x1 + .381966 * (x0 - x1), x1, x2
                y1 = desc_eval(params, dfdp, fnl, i, x1)
        if y1 < y2:
            xbest = x1
            ybest = y1
        else:
            xbest = x2
            ybest = y2
        for ix in range(len(params)):
            params[ix] += xbest * dfdp[ix]
        print '%', params, xbest, ybest
        sys.stdout.flush()
    return params

def mk_mvc_fnl(th0, th1):
    def fnl(params, i, do_print = False):
        k, k1, k2, k3, C = params
        ks, cost, x, y, th = run_mvc(k, k1, k2, k3, C, 100)
        cost *= hypot(y, x) ** 3
        actual_th0 = atan2(y, x)
        actual_th1 = th - actual_th0
        if do_print: print '%', x, y, actual_th0, actual_th1, cost
        err = (th0 - actual_th0) ** 2 + (th1 - actual_th1) ** 2
        multiplier = 1000
        return cost + err * multiplier
    return fnl

def mk_mec_fnl(th0, th1):
    def fnl(params, i, do_print = False):
        k, lam1, lam2 = params
        ks, cost, x, y, th = run_mec_cos(k, lam1, lam2)
        cost *= hypot(y, x)
        actual_th0 = atan2(y, x)
        actual_th1 = th - actual_th0
        if do_print: print '%', x, y, actual_th0, actual_th1, cost
        err = (th0 - actual_th0) ** 2 + (th1 - actual_th1) ** 2
        multiplier = 10
        return cost + err * multiplier
    return fnl

#ks, cost, x, y, th = run_mvc(0, 10, -10, 10, 200)
#print '%', cost, x, y
#print 'stroke showpage'

def mvc_test():
    fnl = mk_mvc_fnl(-pi, pi/4)
    params = [0, 0, 0, 0, 0]
    params = descend3(params, fnl)
    k, k1, k2, k3, C = params
    run_mvc(k, k1, k2, k3, C, 100, True)
    print 'stroke showpage'
    print '%', params

def mec_test():
    th0, th1 = pi/2, pi/2
    fnl = mk_mec_fnl(th0, th1)
    params = [0, 0, 0]
    params = descend2(params, fnl)
    k, lam1, lam2 = params
    run_mec_cos(k, lam1, lam2, 1000, True)
    print 'stroke showpage'
    print '%', params

mvc_test()



# Solver based on direct Newton solving of 4 parameters for each curve
# segment

import sys
from math import *

from Numeric import *
import LinearAlgebra as la

import poly3
import band

class Seg:
    def __init__(self, chord, th):
        self.ks = [0., 0., 0., 0.]
        self.chord = chord
        self.th = th
    def compute_ends(self, ks):
        chord, ch_th = poly3.integ_chord(ks)
        l = chord / self.chord
        thl = ch_th - (-.5 * ks[0] + .125 * ks[1] - 1./48 * ks[2] + 1./384 * ks[3])
        thr = (.5 * ks[0] + .125 * ks[1] + 1./48 * ks[2] + 1./384 * ks[3]) - ch_th
        k0l = l * (ks[0] - .5 * ks[1] + .125 * ks[2] - 1./48 * ks[3])
        k0r = l * (ks[0] + .5 * ks[1] + .125 * ks[2] + 1./48 * ks[3])
        l2 = l * l
        k1l = l2 * (ks[1] - .5 * ks[2] + .125 * ks[3])
        k1r = l2 * (ks[1] + .5 * ks[2] + .125 * ks[3])
        l3 = l2 * l
        k2l = l3 * (ks[2] - .5 * ks[3])
        k2r = l3 * (ks[2] + .5 * ks[3])
        return (thl, k0l, k1l, k2l), (thr, k0r, k1r, k2r), l
    def set_ends_from_ks(self):
        self.endl, self.endr, self.l = self.compute_ends(self.ks)
    def fast_pderivs(self):
        l = self.l
        l2 = l * l
        l3 = l2 * l
        return [((.5, l, 0, 0), (.5, l, 0, 0)),
                ((-1./12, -l/2, l2, 0), (1./12, l/2, l2, 0)),
                ((1./48, l/8, -l2/2, l3), (1./48, l/8, l2/2, l3)),
                ((-1./480, -l/48, l2/8, -l3/2), (1./480, l/48, l2/8, l3/2))]
    def compute_pderivs(self):
        rd = 2e6
        delta = 1./rd
        base_ks = self.ks
        base_endl, base_endr, dummy = self.compute_ends(base_ks)
        result = []
        for i in range(4):
            try_ks = base_ks[:]
            try_ks[i] += delta
            try_endl, try_endr, dummy = self.compute_ends(try_ks)
            deriv_l = (rd * (try_endl[0] - base_endl[0]),
                       rd * (try_endl[1] - base_endl[1]),
                       rd * (try_endl[2] - base_endl[2]),
                       rd * (try_endl[3] - base_endl[3]))
            deriv_r = (rd * (try_endr[0] - base_endr[0]),
                       rd * (try_endr[1] - base_endr[1]),
                       rd * (try_endr[2] - base_endr[2]),
                       rd * (try_endr[3] - base_endr[3]))
            result.append((deriv_l, deriv_r))
        return result

class Node:
    def __init__(self, x, y, ty, th):
        self.x = x
        self.y = y
        self.ty = ty
        self.th = th
    def continuity(self):
        if self.ty == 'o':
            return 4
        elif self.ty in ('c', '[', ']'):
            return 2
        else:
            return 0

def mod_2pi(th):
    u = th / (2 * pi)
    return 2 * pi * (u - floor(u + 0.5))

def setup_path(path):
    segs = []
    nodes = []
    nsegs = len(path)
    if path[0][2] == '{':
        nsegs -= 1
    for i in range(nsegs):
        i1 = (i + 1) % len(path)
        x0, y0, t0 = path[i]
        x1, y1, t1 = path[i1]
        s = Seg(hypot(y1 - y0, x1 - x0), atan2(y1 - y0, x1 - x0))
        segs.append(s)
    for i in range(len(path)):
        x0, y0, t0 = path[i]

        if t0 in ('{', '}', 'v'):
            th = 0
        else:
            s0 = segs[(i + len(path) - 1) % len(path)]
            s1 = segs[i]
            th = mod_2pi(s1.th - s0.th)

        n = Node(x0, y0, t0, th)
        nodes.append(n)
    return segs, nodes

def count_vec(nodes):
    jincs = []
    n = 0
    for i in range(len(nodes)):
        i1 = (i + 1) % len(nodes)
        t0 = nodes[i].ty
        t1 = nodes[i1].ty
        if t0 in ('{', '}', 'v', '[') and t1 in ('{', '}', 'v', ']'):
            jinc = 0
        elif t0 in ('{', '}', 'v', '[') and t1 == 'c':
            jinc = 1
        elif t0 == 'c' and t1 in ('{', '}', 'v', ']'):
            jinc = 1
        elif t0 == 'c' and t1 == 'c':
            jinc = 2
        else:
            jinc = 4
        jincs.append(jinc)
        n += jinc
    return n, jincs

thscale, k0scale, k1scale, k2scale = 1, 1, 1, 1

def inversedot_woodbury(m, v):
    a = zeros((n, 11), Float)
    for i in range(n):
        for j in range(max(-7, -i), min(4, n - i)):
            a[i, j + 7] = m[i, i + j]
    print a
    al, indx, d = band.bandec(a, 7, 3)
    VtZ = identity(4, Float)
    Z = zeros((n, 4), Float)
    for i in range(4):
        u = zeros(n, Float)
        for j in range(4):
            u[j] = m[j, n - 4 + i]
        band.banbks(a, 7, 3, al, indx, u)
        for k in range(n):
            Z[k, i] = u[k]
        #Z[:,i] = u
        for j in range(4):
            VtZ[j, i] += u[n - 4 + j]
    print Z
    print VtZ
    H = la.inverse(VtZ)
    print H
    band.banbks(a, 7, 3, al, indx, v)
    return(v - dot(Z, dot(H, v[n - 4:])))

def inversedot(m, v):
    return dot(la.inverse(m), v)
    n, nn = m.shape
    if 1:
        for i in range(n):
            sys.stdout.write('% ')
            for j in range(n):
                if m[i, j] > 0: sys.stdout.write('+ ')
                elif m[i, j] < 0: sys.stdout.write('- ')
                else: sys.stdout.write('  ')
            sys.stdout.write('\n')

    cyclic = False
    for i in range(4):
        for j in range(n - 4, n):
            if m[i, j] != 0:
                cyclic = True
    print '% cyclic:', cyclic
    if not cyclic:
        a = zeros((n, 11), Float)
        for i in range(n):
            for j in range(max(-5, -i), min(6, n - i)):
                a[i, j + 5] = m[i, i + j]
        for i in range(n):
            sys.stdout.write('% ')
            for j in range(11):
                if a[i, j] > 0: sys.stdout.write('+ ')
                elif a[i, j] < 0: sys.stdout.write('- ')
                else: sys.stdout.write('  ')
            sys.stdout.write('\n')
        al, indx, d = band.bandec(a, 5, 5)
        print a
        band.banbks(a, 5, 5, al, indx, v)
        return v
    else:
        #return inversedot_woodbury(m, v)
        bign = 3 * n
        a = zeros((bign, 11), Float)
        u = zeros(bign, Float)
        for i in range(bign):
            u[i] = v[i % n]
            for j in range(-7, 4):
                a[i, j + 7] = m[i % n, (i + j + 7 * n) % n]
        #print a
        if 1:
            for i in range(bign):
                sys.stdout.write('% ')
                for j in range(11):
                    if a[i, j] > 0: sys.stdout.write('+ ')
                    elif a[i, j] < 0: sys.stdout.write('- ')
                    else: sys.stdout.write('  ')
                sys.stdout.write('\n')
        #print u
        al, indx, d = band.bandec(a, 5, 5)
        band.banbks(a, 5, 5, al, indx, u)
        #print u
        return u[n + 2: 2 * n + 2]

def iter(segs, nodes):
    n, jincs = count_vec(nodes)
    print '%', jincs
    v = zeros(n, Float)
    m = zeros((n, n), Float)
    for i in range(len(segs)):
        segs[i].set_ends_from_ks()
    j = 0
    j0 = 0
    for i in range(len(segs)):
        i1 = (i + 1) % len(nodes)
        t0 = nodes[i].ty
        t1 = nodes[i1].ty
        seg = segs[i]

        derivs = seg.compute_pderivs()
        print '%derivs:', derivs

        jinc = jincs[i]   # the number of params on this seg
        print '%', t0, t1, jinc, j0

        # The constraints are laid out as follows:
        #   constraints that cross the node on the left
        #   constraints on the left side
        #   constraints on the right side
        #   constraints that cross the node on the right

        jj = j0    # the index into the constraint row we're writing
        jthl, jk0l, jk1l, jk2l = -1, -1, -1, -1
        jthr, jk0r, jk1r, jk2r = -1, -1, -1, -1

        # constraints crossing left

        if t0 == 'o':
            jthl = jj + 0
            jk0l = jj + 1
            jk1l = jj + 2
            jk2l = jj + 3
            jj += 4
        elif t0 in ('c', '[', ']'):
            jthl = jj + 0
            jk0l = jj + 1
            jj += 2

        # constraints on left

        if t0 in ('[', 'v', '{') and jinc == 4:
            jk1l = jj
            jj += 1
        if t0 in ('[', 'v', '{', 'c') and jinc == 4:
            jk2l = jj
            jj += 1

        # constraints on right

        if t1 in (']', 'v', '}') and jinc == 4:
            jk1r = jj
            jj += 1
        if t1 in (']', 'v', '}', 'c') and jinc == 4:
            jk2r = jj
            jj += 1

        # constraints crossing right

        jj %= n
        j1 = jj

        if t1 == 'o':
            jthr = jj + 0
            jk0r = jj + 1
            jk1r = jj + 2
            jk2r = jj + 3
            jj += 4
        elif t1 in ('c', '[', ']'):
            jthr = jj + 0
            jk0r = jj + 1
            jj += 2

        print '%', jthl, jk0l, jk1l, jk2l, jthr, jk0r, jk1r, jk2r

        if jthl >= 0:
            v[jthl] += thscale * (nodes[i].th - seg.endl[0])
            if jinc == 1:
                m[jthl][j] += derivs[0][0][0]
            elif jinc == 2:
                m[jthl][j + 1] += derivs[0][0][0]
                m[jthl][j] += derivs[1][0][0]
            elif jinc == 4:
                m[jthl][j + 2] += derivs[0][0][0]
                m[jthl][j + 3] += derivs[1][0][0]
                m[jthl][j + 0] += derivs[2][0][0]
                m[jthl][j + 1] += derivs[3][0][0]
        if jk0l >= 0:
            v[jk0l] += k0scale * seg.endl[1]
            if jinc == 1:
                m[jk0l][j] -= derivs[0][0][1]
            elif jinc == 2:
                m[jk0l][j + 1] -= derivs[0][0][1]
                m[jk0l][j] -= derivs[1][0][1]
            elif jinc == 4:
                m[jk0l][j + 2] -= derivs[0][0][1]
                m[jk0l][j + 3] -= derivs[1][0][1]
                m[jk0l][j + 0] -= derivs[2][0][1]
                m[jk0l][j + 1] -= derivs[3][0][1]
        if jk1l >= 0:
            v[jk1l] += k1scale * seg.endl[2]
            m[jk1l][j + 2] -= derivs[0][0][2]
            m[jk1l][j + 3] -= derivs[1][0][2]
            m[jk1l][j + 0] -= derivs[2][0][2]
            m[jk1l][j + 1] -= derivs[3][0][2]
        if jk2l >= 0:
            v[jk2l] += k2scale * seg.endl[3]
            m[jk2l][j + 2] -= derivs[0][0][3]
            m[jk2l][j + 3] -= derivs[1][0][3]
            m[jk2l][j + 0] -= derivs[2][0][3]
            m[jk2l][j + 1] -= derivs[3][0][3]

        if jthr >= 0:
            v[jthr] -= thscale * seg.endr[0]
            if jinc == 1:
                m[jthr][j] += derivs[0][1][0]
            elif jinc == 2:
                m[jthr][j + 1] += derivs[0][1][0]
                m[jthr][j + 0] += derivs[1][1][0]
            elif jinc == 4:
                m[jthr][j + 2] += derivs[0][1][0]
                m[jthr][j + 3] += derivs[1][1][0]
                m[jthr][j + 0] += derivs[2][1][0]
                m[jthr][j + 1] += derivs[3][1][0]
        if jk0r >= 0:
            v[jk0r] -= k0scale * seg.endr[1]
            if jinc == 1:
                m[jk0r][j] += derivs[0][1][1]
            elif jinc == 2:
                m[jk0r][j + 1] += derivs[0][1][1]
                m[jk0r][j + 0] += derivs[1][1][1]
            elif jinc == 4:
                m[jk0r][j + 2] += derivs[0][1][1]
                m[jk0r][j + 3] += derivs[1][1][1]
                m[jk0r][j + 0] += derivs[2][1][1]
                m[jk0r][j + 1] += derivs[3][1][1]
        if jk1r >= 0:
            v[jk1r] -= k1scale * seg.endr[2]
            m[jk1r][j + 2] += derivs[0][1][2]
            m[jk1r][j + 3] += derivs[1][1][2]
            m[jk1r][j + 0] += derivs[2][1][2]
            m[jk1r][j + 1] += derivs[3][1][2]
        if jk2r >= 0:
            v[jk2r] -= k2scale * seg.endr[3]
            m[jk2r][j + 2] += derivs[0][1][3]
            m[jk2r][j + 3] += derivs[1][1][3]
            m[jk2r][j + 0] += derivs[2][1][3]
            m[jk2r][j + 1] += derivs[3][1][3]

        j += jinc
        j0 = j1
    #print m
    dk = inversedot(m, v)
    dkmul = 1
    j = 0
    for i in range(len(segs)):
        jinc = jincs[i]
        if jinc == 1:
            segs[i].ks[0] += dkmul * dk[j]
        elif jinc == 2:
            segs[i].ks[0] += dkmul * dk[j + 1]
            segs[i].ks[1] += dkmul * dk[j + 0]
        elif jinc == 4:
            segs[i].ks[0] += dkmul * dk[j + 2]
            segs[i].ks[1] += dkmul * dk[j + 3]
            segs[i].ks[2] += dkmul * dk[j + 0]
            segs[i].ks[3] += dkmul * dk[j + 1]
        j += jinc

    norm = 0.
    for i in range(len(dk)):
        norm += dk[i] * dk[i]
    return sqrt(norm)


def plot_path(segs, nodes, tol = 1.0, show_cpts = False):
    if show_cpts:
        cpts = []
    j = 0
    cmd = 'moveto'
    for i in range(len(segs)):
        i1 = (i + 1) % len(nodes)
        n0 = nodes[i]
        n1 = nodes[i1]
        x0, y0, t0 = n0.x, n0.y, n0.ty
        x1, y1, t1 = n1.x, n1.y, n1.ty
        ks = segs[i].ks
        abs_ks = abs(ks[0]) + abs(ks[1] / 2) + abs(ks[2] / 8) + abs(ks[3] / 48)
        n_subdiv = int(ceil(.001 + tol * abs_ks))
        n_subhalf = (n_subdiv + 1) / 2
        if n_subdiv > 1:
            n_subdiv = n_subhalf * 2
        ksp = (ks[0] * .5, ks[1] * .25, ks[2] * .125, ks[3] * .0625)
        pside = poly3.int_3spiro_poly(ksp, n_subhalf)
        ksm = (ks[0] * -.5, ks[1] * .25, ks[2] * -.125, ks[3] * .0625)
        mside = poly3.int_3spiro_poly(ksm, n_subhalf)
        mside.reverse()
        for j in range(len(mside)):
            mside[j] = (-mside[j][0], -mside[j][1])
        if n_subdiv > 1:
            pts = mside + pside[1:]
        else:
            pts = mside[:1] + pside[1:]
        chord_th = atan2(y1 - y0, x1 - x0)
        chord_len = hypot(y1 - y0, x1 - x0)
        rot = chord_th - atan2(pts[-1][1] - pts[0][1], pts[-1][0] - pts[0][0])
        scale = chord_len / hypot(pts[-1][1] - pts[0][1], pts[-1][0] - pts[0][0])
        u, v = scale * cos(rot), scale * sin(rot)
        xt = x0 - u * pts[0][0] + v * pts[0][1]
        yt = y0 - u * pts[0][1] - v * pts[0][0]
        s = -.5
        for x, y in pts:
            xp, yp = xt + u * x - v * y, yt + u * y + v * x
            thp = (((ks[3]/24 * s + ks[2]/6) * s + ks[1] / 2) * s + ks[0]) * s + rot
            up, vp = scale / (1.5 * n_subdiv) * cos(thp), scale / (1.5 * n_subdiv) * sin(thp)
            if s == -.5:
                if cmd == 'moveto':
                    print xp, yp, 'moveto'
                    cmd = 'curveto'
            else:
                if show_cpts:
                    cpts.append((xlast + ulast, ylast + vlast))
                    cpts.append((xp - up, yp - vp))
                print xlast + ulast, ylast + vlast, xp - up, yp - vp, xp, yp, 'curveto'
            xlast, ylast, ulast, vlast = xp, yp, up, vp
            s += 1. / n_subdiv
        if t1 == 'v':
            j += 2
        else:
            j += 1
    print 'stroke'
    if show_cpts:
        for x, y in cpts:
            print 'gsave 0 0 1 setrgbcolor', x, y, 'translate circle fill grestore'

def plot_ks(segs, nodes, xo, yo, xscale, yscale):
    j = 0
    cmd = 'moveto'
    x = xo
    for i in range(len(segs)):
        i1 = (i + 1) % len(nodes)
        n0 = nodes[i]
        n1 = nodes[i1]
        x0, y0, t0 = n0.x, n0.y, n0.ty
        x1, y1, t1 = n1.x, n1.y, n1.ty
        ks = segs[i].ks
        chord, ch_th = poly3.integ_chord(ks)
        l = chord/segs[i].chord
        k0 = l * poly3.eval_cubic(ks[0], ks[1], .5 * ks[2], 1./6 * ks[3], -.5)
        print x, yo + yscale * k0, cmd
        cmd = 'lineto'
        k3 = l * poly3.eval_cubic(ks[0], ks[1], .5 * ks[2], 1./6 * ks[3], .5)
        k1 = k0 + l/3 * (ks[1] - 0.5 * ks[2] + .125 * ks[3])
        k2 = k3 - l/3 * (ks[1] + 0.5 * ks[2] + .125 * ks[3])
        print x + xscale / l / 3., yo + yscale * k1
        print x + 2 * xscale / l / 3., yo + yscale * k2
        print x + xscale / l, yo + yscale * k3, 'curveto'
        x += xscale / l
        if t1 == 'v':
            j += 2
        else:
            j += 1
    print 'stroke'
    print xo, yo, 'moveto', x, yo, 'lineto stroke'

def plot_nodes(nodes, segs):
    for i in range(len(nodes)):
        n = nodes[i]
        print 'gsave', n.x, n.y, 'translate'
        if n.ty in ('c', '[', ']'):
            th = segs[i].th - segs[i].endl[0]
            if n.ty == ']': th += pi
            print 180 * th / pi, 'rotate'
        if n.ty == 'o':
            print 'circle fill'
        elif n.ty == 'c':
            print '3 4 poly fill'
        elif n.ty in ('v', '{', '}'):
            print '45 rotate 3 4 poly fill'
        elif n.ty in ('[', ']'):
            print '0 -3 moveto 0 0 3 90 270 arc fill'
        else:
            print '5 3 poly fill'
        print 'grestore'

def prologue():
    print '/ss 2 def'
    print '/circle { ss 0 moveto currentpoint exch ss sub exch ss 0 360 arc } bind def'
    print '/poly {'
    print '    dup false exch {'
    print '        0 3 index 2 index { lineto } { moveto } ifelse pop true'
    print '        360.0 2 index div rotate'
    print '    } repeat pop pop pop'
    print '} bind def'

def run_path(path, show_iter = False, n = 10, xo = 36, yo = 550, xscale = .25, yscale = 2000, pl_nodes = True):
    segs, nodes = setup_path(path)
    print '.5 setlinewidth'
    for i in range(n):
        if i == n - 1:
            print '0 0 0 setrgbcolor 1 setlinewidth'
        elif i == 0:
            print '1 0 0 setrgbcolor'
        elif i == 1:
            print '0 0.5 0 setrgbcolor'
        elif i == 2:
            print '0.3 0.3 0.3 setrgbcolor'
        norm = iter(segs, nodes)
        print '% norm =', norm
        if show_iter or i == n - 1:
            #print '1 0 0 setrgbcolor'
            #plot_path(segs, nodes, tol = 5)
            #print '0 0 0 setrgbcolor'
            plot_path(segs, nodes, tol = 1.0)
            plot_ks(segs, nodes, xo, yo, xscale, yscale)
    if pl_nodes:
        plot_nodes(nodes, segs)

if __name__ == '__main__':
    if 1:
        path = [(100, 350, 'o'), (225, 350, 'o'), (350, 450, 'o'),
                (450, 400, 'o'), (315, 205, 'o'), (300, 200, 'o'),
                (285, 205, 'o')]

    if 1:
        path = [(100, 350, 'o'), (175, 375, '['), (250, 375, ']'), (325, 425, '['),
                (325, 450, ']'),
                (400, 475, 'o'), (350, 200, 'c')]

    if 0:
        ecc, ty, ty1 = 0.56199, 'c', 'c'
        ecc, ty, ty1 = 0.49076, 'o', 'o',
        ecc, ty, ty1 = 0.42637, 'o', 'c'
        path = [(300 - 200 * ecc, 300, ty), (300, 100, ty1),
                (300 + 200 * ecc, 300, ty), (300, 500, ty1)]

    # difficult path #3
    if 0:
        path = [(100, 300, '{'), (225, 350, 'o'), (350, 500, 'o'),
                (450, 500, 'o'), (450, 450, 'o'), (300, 200, '}')]

    if 0:
        path = [(100, 100, '{'), (200, 180, 'v'), (250, 215, 'o'),
                (300, 200, 'o'), (400, 100, '}')]

    if 0:
        praw = [
        (134, 90, 'o'),
        (192, 68, 'o'),
        (246, 66, 'o'),
        (307, 107, 'o'),
        (314, 154, '['),
        (317, 323, ']'),
        (347, 389, 'o'),
        (373, 379, 'v'),
        (386, 391, 'v'),
        (365, 409, 'o'),
        (335, 419, 'o'),
        (273, 390, 'v'),
        (251, 405, 'o'),
        (203, 423, 'o'),
        (102, 387, 'o'),
        (94, 321, 'o'),
        (143, 276, 'o'),
        (230, 251, 'o'),
        (260, 250, 'v'),
        (260, 220, '['),
        (258, 157, ']'),
        (243, 110, 'o'),
        (159, 99, 'o'),
        (141, 121, 'o'),
        (156, 158, 'o'),
        (121, 184, 'o'),
        (106, 117, 'o')]
        if 0:
            praw = [
            (275, 56, 'o'),
            (291, 75, 'v'),
            (312, 61, 'o'),
            (344, 50, 'o'),
            (373, 72, 'o'),
            (356, 91, 'o'),
            (334, 81, 'o'),
            (297, 85, 'v'),
            (306, 116, 'o'),
            (287, 180, 'o'),
            (236, 213, 'o'),
            (182, 212, 'o'),
            (157, 201, 'v'),
            (149, 209, 'o'),
            (143, 230, 'o'),
            (162, 246, 'c'),
            (202, 252, 'o'),
            (299, 260, 'o'),
            (331, 282, 'o'),
            (341, 341, 'o'),
            (308, 390, 'o'),
            (258, 417, 'o'),
            (185, 422, 'o'),
            (106, 377, 'o'),
            (110, 325, 'o'),
            (133, 296, 'o'),
            (147, 283, 'v'),
            (117, 238, 'o'),
            (133, 205, 'o'),
            (147, 191, 'v'),
            (126, 159, 'o'),
            (128, 94, 'o'),
            (167, 50, 'o'),
            (237, 39, 'o')]
            
        path = []
        for x, y, ty in praw:
            #if ty == 'o': ty = 'c'
            path.append((x, 550 - y, ty))

    if 0:
        path = [(100, 300, 'o'), (300, 100, 'o'), (300, 500, 'o')]

    if 0:
        # Woodford data set
        ty = 'o'
        praw = [(0, 0, '{'), (1, 1.9, ty), (2, 2.7, ty), (3, 2.6, ty),
                (4, 1.6, ty), (5, .89, ty), (6, 1.2, '}')]
        path = []
        for x, y, t in praw:
            path.append((100 + 80 * x, 100 + 80 * y, t))

    if 0:
        ycen = 32
        yrise = 0
        path = []
        ty = '{'
        for i in range(10):
            path.append((50 + i * 30, 250 + (10 - i) * yrise, ty))
            ty = 'o'
        path.append((350, 250 + ycen, ty))
        for i in range(1, 10):
            path.append((350 + i * 30, 250 + i * yrise, ty))
        path.append((650, 250 + 10 * yrise, '}'))

    prologue()

    run_path(path, show_iter = True, n=5)
